//@author : Debasish Das ddas@northwestern.edu

//Dbrent Method used from Numerical Recipes

//Most sizing problems which solve dual of the problem solves a variant of 
//Lagrangian Subproblem. Generally a simple cost function is assumed which
//based on Chen et al. yields a solution in closed form. I believe that's
//one of the reasons of the efficiency of Chen et al's algorithm. In this code
//we explore LRS behavior with a complex cost function
#ifndef COMPLEX_COST_VARS_H
#define COMPLEX_COST_VARS_H

#include "../graph/graph_utilities.h"
#include "nr3.h"
#include "sizing_vars.h"
#include "pmatrix.h"
#include "sparse.h"

inline double 
power(double x, int p) {
  double ret = 1.0;
  if ( p > 0 ) {
    do {
      ret *= x;
      p--;
    } while (p != 0);
  } else if ( p < 0 ){
    do{
      ret /= x;
      p++;
    } while (p != 0);
  } else {
    ret = 1.0;
  }
  return ret;
}

struct Bracketmethod {
  Doub ax,bx,cx,fa,fb,fc;
  template <class T>
  void bracket(const Doub a, const Doub b, T &func)
  {
    const Doub GOLD=1.618034,GLIMIT=100.0,TINY=1.0e-20;
    ax=a; bx=b;
    Doub fu;
    fa=func(ax);
    fb=func(bx);
    if (fb > fa) {
      SWAP(ax,bx);
      SWAP(fb,fa);
    }
    cx=bx+GOLD*(bx-ax);
    fc=func(cx);
    while (fb > fc) {
      Doub r=(bx-ax)*(fb-fc);
      Doub q=(bx-cx)*(fb-fa);
      Doub u=bx-((bx-cx)*q-(bx-ax)*r)/
	(2.0*SIGN(MAX(abs(q-r),TINY),q-r));
      Doub ulim=bx+GLIMIT*(cx-bx);
      if ((bx-u)*(u-cx) > 0.0) {
	fu=func(u);
	if (fu < fc) {
	  ax=bx;
	  bx=u;
	  fa=fb;
	  fb=fu;
	  return;
	} else if (fu > fb) {
	  cx=u;
	  fc=fu;
	  return;
	}
	u=cx+GOLD*(cx-bx);
	fu=func(u);
      } else if ((cx-u)*(u-ulim) > 0.0) {
	fu=func(u);
	if (fu < fc) {
	  shft3(bx,cx,u,u+GOLD*(u-cx));
	  shft3(fb,fc,fu,func(u));
	}
      } else if ((u-ulim)*(ulim-cx) >= 0.0) {
	u=ulim;
	fu=func(u);
      } else {
	u=cx+GOLD*(cx-bx);
	fu=func(u);
      }
      shft3(ax,bx,cx,u);
      shft3(fa,fb,fc,fu);
    }
  }
  inline void shft2(Doub &a, Doub &b, const Doub c)
  {
    a=b;
    b=c;
  }
  inline void shft3(Doub &a, Doub &b, Doub &c, const Doub d)
  {
    a=b;
    b=c;
    c=d;
  }
  inline void mov3(Doub &a, Doub &b, Doub &c, const Doub d, const Doub e,
		   const Doub f)
  {
    a=d; b=e; c=f;
  }
};

struct Dbrent : Bracketmethod {
  Doub xmin,fmin;
  const Doub tol;
  Dbrent(const Doub toll=3.0e-8) : tol(toll) {}
  template <class T>
  Doub minimize(T &funcd)
  {
    const Int ITMAX=100;
    const Doub ZEPS=numeric_limits<Doub>::epsilon()*1.0e-3;
    Bool ok1,ok2;
    Doub a,b,d=0.0,d1,d2,du,dv,dw,dx,e=0.0;
    Doub fu,fv,fw,fx,olde,tol1,tol2,u,u1,u2,v,w,x,xm;
    
    a=(ax < cx ? ax : cx);
    b=(ax > cx ? ax : cx);
    x=w=v=bx;
    fw=fv=fx=funcd(x);
    dw=dv=dx=funcd.df(x);
    for (Int iter=0;iter<ITMAX;iter++) {
      xm=0.5*(a+b);
      tol1=tol*abs(x)+ZEPS;
      tol2=2.0*tol1;
      if (abs(x-xm) <= (tol2-0.5*(b-a))) {
	fmin=fx;
	return xmin=x;
      }
      if (abs(e) > tol1) {
	d1=2.0*(b-a);
	d2=d1;
	if (dw != dx) d1=(w-x)*dx/(dx-dw);
	if (dv != dx) d2=(v-x)*dx/(dx-dv);
	u1=x+d1;
	u2=x+d2;
	ok1 = (a-u1)*(u1-b) > 0.0 && dx*d1 <= 0.0;
	ok2 = (a-u2)*(u2-b) > 0.0 && dx*d2 <= 0.0;
	olde=e;
	e=d;
	if (ok1 || ok2) {
	  if (ok1 && ok2)
	    d=(abs(d1) < abs(d2) ? d1 : d2);
	  else if (ok1)
	    d=d1;
	  else
	    d=d2;
	  if (abs(d) <= abs(0.5*olde)) {
	    u=x+d;
	    if (u-a < tol2 || b-u < tol2)
	      d=SIGN(tol1,xm-x);
	  } else {
	    d=0.5*(e=(dx >= 0.0 ? a-x : b-x));
	  }
	} else {
	  d=0.5*(e=(dx >= 0.0 ? a-x : b-x));
	}
      } else {
	d=0.5*(e=(dx >= 0.0 ? a-x : b-x));
      }
      if (abs(d) >= tol1) {
	u=x+d;
	fu=funcd(u);
      } else {
	u=x+SIGN(tol1,d);
	fu=funcd(u);
	if (fu > fx) {
	  fmin=fx;
	  return xmin=x;
	}
      }
      du=funcd.df(u);
      if (fu <= fx) {
	if (u >= x) a=x; else b=x;
	mov3(v,fv,dv,w,fw,dw);
	mov3(w,fw,dw,x,fx,dx);
	mov3(x,fx,dx,u,fu,du);
      } else {
	if (u < x) a=u; else b=u;
	if (fu <= fw || w == x) {
	  mov3(v,fv,dv,w,fw,dw);
	  mov3(w,fw,dw,u,fu,du);
	} else if (fu < fv || v == x || v == w) {
	  mov3(v,fv,dv,u,fu,du);
	}
      }
    }
    fprintf(stdout, "Too many iterations in routine dbrent\n");
    return -1.0;
  }
};

template <class T>
Doub rtnewt(T &funcd, const Doub x1, const Doub x2, const Doub xacc) 
{
  const Int JMAX=20;
  Doub rtn=0.5*(x1+x2);
  for (Int j=0;j<JMAX;j++) {
    Doub f=funcd(rtn);
    Doub df=funcd.df(rtn);
    Doub dx=f/df;
    rtn -= dx;
    if ((x1-rtn)*(rtn-x2) < 0.0) {
      fprintf(stdout, "Jumped out of brackets in rtnewt\n");
      return -1.0;
    }
    if (abs(dx) < xacc) return rtn;
  }
  fprintf(stdout, "Maximum number of iterations exceeded in rtnewt\n");
  return -1.0;
}

template <class T>
Doub rtsafe(T &funcd, const Doub x1, const Doub x2, const Doub xacc) 
{
  const Int MAXIT=100;
  Doub xh,xl;
  Doub fl=funcd(x1);
  Doub fh=funcd(x2);
  if ((fl > 0.0 && fh > 0.0) || (fl < 0.0 && fh < 0.0))
    throw("Root must be bracketed in rtsafe");
  if (fl == 0.0) return x1;
  if (fh == 0.0) return x2;
  if (fl < 0.0) {
    xl=x1;
    xh=x2;
  } else {
    xh=x1;
    xl=x2;
  }
  Doub rts=0.5*(x1+x2);
  Doub dxold=abs(x2-x1);
  Doub dx=dxold;
  Doub f=funcd(rts);
  Doub df=funcd.df(rts);
  for (Int j=0;j<MAXIT;j++) {
    if ((((rts-xh)*df-f)*((rts-xl)*df-f) > 0.0)
	|| (abs(2.0*f) > abs(dxold*df))) {
      dxold=dx;
      dx=0.5*(xh-xl);
      rts=xl+dx;
      if (xl == rts) return rts;
    } else {
      dxold=dx;
      dx=f/df;
      Doub temp=rts;
      rts -= dx;
      if (temp == rts) return rts;
    }
    if (abs(dx) < xacc) return rts;
    Doub f=funcd(rts);
    //Doub df=funcd.df(rts);
    if (f < 0.0)
      xl=rts;
    else
      xh=rts;
  }
  fprintf(stdout, "Maximum number of iterations exceeded in rtsafe\n");
  return -1.0;
}

class complex_cost_vars : public sizing_vars
{
 public:
  struct func{
    std::vector<double> coeffs;
    
    func(size_t n) {
      coeffs.resize(n, 0.0);
    }
    
    func(std::vector<double> c){
      coeffs = c;
    }
    
    double operator() (const double x)
    {
      double ret = 0.0;
      for(size_t i = 0; i < coeffs.size(); i++)
	ret += coeffs[i]*power(x,i);
      return ret;
    }
    
    double df(const double x)
    {
      double ret = 0.0, cterm = 0.0, pterm = 0.0;
      for(size_t i = 0; i < coeffs.size(); i++){
	double c = coeffs[i];
	cterm = c*i;
	if ( cterm == 0 ) continue;
	else pterm = power(x,i-1);
	ret += cterm*pterm;
      }
      return ret;
    }
    
    void resize(size_t n) {
      coeffs.resize(n);
    }
  };
  
  struct fast_polynomial{
    typedef std::vector<double> cvec;
    typedef cvec::iterator cvec_iter;
    
    cvec pos_coeffs;
    cvec neg_coeffs;
    int order;
    int min_p;
    int max_p;
    
    fast_polynomial() {
      min_p = 0;
      max_p = 0;
      order = max_p - min_p;
    }
    
    int get_min_power() {
      return min_p;
    }
    
    int get_max_power() {
      return max_p;
    }
    
    void clear() {
      pos_coeffs.clear();
      neg_coeffs.clear();
    }
    
    double operator() (const double x) {
      double ret = 0.0;
      for(size_t i = 0; i < pos_coeffs.size(); i++){
	double c = pos_coeffs[i];
	ret += c*power(x, i);
      }
      for(size_t i = 0; i < neg_coeffs.size(); i++){
	double c = neg_coeffs[i];
	int p = -i - 1;
	ret += c*power(x, p);
      }
      return ret;
    }
    
    double df(const double x) {
      double cterm = 0.0, pterm = 0.0, ret = 0.0;
      for(size_t i = 0; i < pos_coeffs.size(); i++){
	double c = pos_coeffs[i];
	cterm = c*i;
	if ( cterm == 0 ) continue;
	else pterm = power(x,i-1);
	ret += cterm*pterm;
      }
      for(size_t i = 0; i < neg_coeffs.size(); i++){
	double c = neg_coeffs[i];
	int p = -i - 1;
	cterm = c*p;
	if ( cterm == 0 ) continue;
	else pterm = power(x,p-1);
	ret += cterm*pterm;
      }
      return ret;
    }
    
    size_t get_order() {
      order = max_p - min_p;
      return order;
    }
    
    void insert(int p, double c) {
      if ( p == 0 ) {
	if ( pos_coeffs.size() > 0 ) 
	  pos_coeffs[p] += c;
	else 
	  pos_coeffs.push_back(c);
      }
      else if ( p > 0) {
	if ( max_p < p ) {
	  for(int i = max_p; i <= p; i++)
	    pos_coeffs.push_back(0.0);
	  max_p = p;
	}
	pos_coeffs[p] += c;
      } else {
	int index = -p - 1;
	if ( index == 0 ) {
	  if ( neg_coeffs.size() > 0 )
	    neg_coeffs[index] += c;
	  else
	    neg_coeffs.push_back(c);
	  min_p = p;
	} else {
	  if ( min_p > p ) {
	    int start = min_p;
	    if ( min_p < 0 ) {
	      start = -min_p - 1;
	    }
	    for(int i = start; i <= index; i++) 
	      neg_coeffs.push_back(0.0);
	    min_p = p;
	  }
	  neg_coeffs[index] += c;
	}
      }
    }
    
    void generate_derivative(complex_cost_vars::fast_polynomial& dp)
    {
      double cterm = 0.0;
      int pterm = 0;
      
      for(size_t i = 0; i < pos_coeffs.size(); i++){
	double c = pos_coeffs[i];
	cterm = c*i;
	if ( cterm == 0 ) continue;
	else {
	  pterm = i-1;
	  dp.insert(pterm, cterm);
	}
      }
      for(size_t i = 0; i < neg_coeffs.size(); i++){
	double c = neg_coeffs[i];
	int p = -i - 1;
	cterm = c*p;
	if ( cterm == 0 ) continue;
	else {
	  pterm = p-1;
	  dp.insert(pterm, cterm);
	}
      }
    }
    
    bool get_coefficient(int i, double &c)
    {
      if ( i > max_p || i < min_p ) {
	c = 0.0;
	return false;
      }
      if ( i >= 0 ){
	c = pos_coeffs[i];
	return true;
      } else {
	size_t index = -i - 1;
	c = neg_coeffs[index];
	return true;
      }
    } 
  };
  
  struct polynomial{
    typedef std::pair<int,double> pc_pair;
    typedef std::map<int,double> pc_map;
    typedef pc_map::iterator pc_map_iter;
    
    pc_map coeffs;
    int order;
    int min_p;
    int max_p;
    
    polynomial() {
      order = -1;
      min_p = 0;
      max_p = 0;
    }
    
    int get_min_power() {
      return min_p;
    }
    
    bool insert(int p, double c) {
      pc_map_iter pc_iter = coeffs.find(p);
      if ( pc_iter != coeffs.end() ) {
	(*pc_iter).second += c;
	return false;
      }
      else {
	coeffs.insert(std::make_pair(p,c));
	return true;
      }
    }
    
    void clear() {
      coeffs.clear();
    }
    
    double operator() (const double x)
    {
      double ret = 0.0;
      for(pc_map_iter pc_iter = coeffs.begin();
	  pc_iter != coeffs.end(); 
	  pc_iter++) 
	{
	  pc_pair pcp = *pc_iter;
	  int p = pcp.first;
	  double c = pcp.second;
	  ret += c*power(x,p);
	}
      return ret;
    }

    double df(const double x)
    {
      double cterm = 0.0, pterm = 0.0, ret = 0.0;
      for(pc_map_iter pc_iter = coeffs.begin();
	  pc_iter != coeffs.end();
	  pc_iter++)
	{
	  pc_pair pcp = *pc_iter;
	  int p = pcp.first;
	  double c = pcp.second;
	  cterm = c*p;
	  if ( cterm == 0 ) continue;
	  else pterm = power(x, p-1);
	  ret += cterm*pterm;
	}
      return ret;
    }
    
    bool is_linear() 
    {
      coeff_map_iter iter = coeffs.begin();
      size_t i = 0;
      int max_power, min_power;
      for(; iter != coeffs.end(); iter++) {
	pc_pair pcp = *iter;
	int p = pcp.first;
	if (i == 0 ) { 
	  max_power = p;
	  min_power = p;
	} else {
	  if ( max_power < p ) max_power = p;
	  if ( min_power > p ) min_power = p;
	}
	i++;
      }
      if ( fabs(max_power)*fabs(min_power) == 0 )
	return true;
      else
	return false;
    }

    size_t get_order()
    {
      if ( order != -1 ) {
	return order;
      } else 
	{
	  bool init = false;
	  int max_power = 0, min_power = 0;
	  coeff_map_iter iter = coeffs.begin();
	  for(; iter != coeffs.end(); iter++) {
	    pc_pair pcp = *iter;
	    int p = pcp.first;
	    if ( !init ) { 
	      max_power = p;
	      min_power = p;
	      init = true;
	    } else {
	      if ( max_power < p ) max_power = p;
	      if ( min_power > p ) min_power = p;
	    }
	  }
	  if ( max_power == 0 && min_power == 0 ) {
	    order = 0;
	    return order;
	  } else {
	    order = max_power - min_power;
	    max_p = max_power, min_p = min_power;
	    return order;
	  }
	}
    }
  };
  
  //vars array in complex cost is maintained by LRS algorithm
  //that's why we use struct for this implementation
  struct complex_cost
  {
    NRsparseMat conductance;
    math::pmatrix Alpha, Beta;
    
    complex_cost() {}
    
    void generate_cost(size_t which_var, double lb, double ub,
		       std::vector<double> vvars,
		       complex_cost_vars::polynomial &p);
    
    double compute_cost(std::vector<double> vvars);

    double compute_cost_dense(std::vector<double> vvars);
    
  };
  
  typedef std::pair<int,double> coeff_pair;
  typedef std::map<int, double> coeff_map;
  typedef coeff_map::iterator coeff_map_iter;
  
 public:
  //ctor
  complex_cost_vars(){
    p_cc = 0x0;
  }
  
  virtual double compute_cost() const;
  double compute_complex_cost() const;
  
  void attach_cost(complex_cost * cc) {
    if (  ! cc ) {
      fprintf(stderr, "Cost is NULL in attach cost\n");
      exit(1);
    }
    p_cc = cc;
  }
 protected:
  complex_cost *p_cc;
  
  //Functions related to variable updates
  virtual void initialize_vars();
  
  int update_var(size_t which_var, double value, double lb, double ub);
  
  int update_var(size_t which_var, complex_cost_vars::fast_polynomial p, double value, 
		 double lb, double ub);
  
  virtual bool LRS_update_var(size_t which_var);
  
  double generate_constraints_bound(size_t which_var);
  
  void generate_polynomial(size_t which_var, complex_cost_vars::fast_polynomial& p);
  
  void generate_coefficients(complex_cost_vars::fast_polynomial p, 
			     std::vector<double>& coeffs);
  
  double solve_cubic(std::vector<double> coeffs, double lb);
  
  double solve_quadratic(std::vector<double> coeffs);
  
};

#endif
