#include "../misc/common.h"
#include "../misc/options.h"
#include "../misc/trace.h"

#include <limits>
#include "complex_cost_vars_map.h"
#include "complex_cost.h"

#include "../misc/timer.h"

using misc::options;

#define TOL 1e-6
RegisterOption(lrs, "analytical", "LRS solver : nr|dbrent|analytical");

os::timer2 t;

void complex_cost_vars::initialize_vars()
{
  assert(sg_ != 0);

  target_ = 0;
  cost_ub_ = 0;
  
  t_cost_eval = 0;
  t_solve_opt = 0;
  
  for(size_t i = 0; i < num_vars(); ++i)
    {
      vars_[i] = sg_->get_var_lb(i);
      
      //Cost for dynamic power
      double w = sg_->get_var_weight(i);
      
      target_ += w*vars_[i];
      
      cost_ub_ += w*sg_->get_var_ub(i);
    }
  
  //Cost for leakage power
#if 1
  double complex_cost = compute_complex_cost();  
  target_ += complex_cost;
  std::vector<double> ub_vec;
  ub_vec.resize(vars_.size());
  for(size_t i = 0; i < vars_.size(); i++) {
    ub_vec[i] = sg_->get_var_ub(i);
  }
  complex_cost = sg_->compute_cost(&ub_vec[0]);
  cost_ub_ += complex_cost;
#endif
}

double complex_cost_vars::compute_cost() const
{
  assert(!use_T_2_);
  assert(sg_ != 0);
  
  double target = std::inner_product(vars_.begin(), vars_.end(), 
				     sg_->get_var_weights(), 0.0);
#if 1
  target += compute_complex_cost();
#endif
  double fd = compute_fd();
  
  return target+fd;
}
 
double complex_cost_vars::compute_complex_cost() const
{
  double cost = sg_->compute_cost(get_vars());
  return cost;
}

int complex_cost_vars::
update_var(size_t which_var, double value, double lb, double ub)
{
  assert(sg_ != 0);
  
  int ret = UPDATED;
  double real = value;
  if (value < lb)
    real = lb, ret = SET_MIN;
  else if (value > ub)
    real = ub, ret = SET_MAX;
  
  if (fabs(real-vars_[which_var]) < TOL)
    return NO_CHANGE;
  
  double w = sg_->get_var_weight(which_var);
  double cost = vars_[which_var]*w;
#if 1
  cost += compute_complex_cost();
#endif
  target_ -= cost;
  vars_[which_var] = real;
  cost = vars_[which_var]*w;
#if 1
  cost += compute_complex_cost();
#endif
  target_ += cost;
  return ret;
}

int complex_cost_vars::
update_var(size_t which_var, complex_cost_vars::polynomial p, double value, double lb, double ub)
{
  assert(sg_ != 0);
  
  int ret = UPDATED;
  double real = value;
  if (value < lb)
    real = lb, ret = SET_MIN;
  else if (value > ub)
    real = ub, ret = SET_MAX;
  
  if (fabs(real-vars_[which_var]) < TOL)
    return NO_CHANGE;
  
  double old_val = vars_[which_var];
  double old_cost = p(old_val);
  target_ -= old_cost;
  vars_[which_var] = real;
  double cost = p(real);
  target_ += cost;
  return ret;
}

bool complex_cost_vars::
LRS_update_var(size_t which_var)
{
  TraceScope("GTSIZING", 10);
  
  //Generate the coefficients of the polynomial P(which_var)
  //n is the order of the polynomial
  complex_cost_vars::polynomial p;
  t_cost_eval = 0;
  t_solve_opt = 0;
  
  t.start_now();
  generate_polynomial(which_var, p);
  t.stop_now();
  t_cost_eval += t.user();
  
  std::vector<double> coeffs;
  t.start_now();
  generate_derivative(p, coeffs);
  t.stop_now();
  t_solve_opt += t.user();
  
  double lb = sg_->get_var_lb(which_var);
  //double ub = generate_constraints_bound(which_var);
  double ub = sg_->get_var_ub(which_var);
  
  double min = 0.0;
  
  switch(coeffs.size() - 1) 
    {
    case 0:
      min = lb;
      break;
    case 1 :
      min =  -coeffs[0]/coeffs[1];
      if ( min < 0 )
	min = lb;
      break;
    case 2:
      min = solve_quadratic(coeffs);
      if ( min < 0 )
	min = lb;
      break;
    case 3:
      std::string lrs_solver = options::get().as_str("lrs");
      if (lrs_solver == "dbrent") {
	//Numerical minimization using Brent's method
	t.start_now();
	Dbrent dbrent;
	dbrent.bracket(0, sg_->get_var_ub(which_var),p);
	min = dbrent.minimize(p);
	t.stop_now();
	t_solve_opt += t.user();
	TracePrint("Minimum computed using Brent : %e\n", min);
      } else if (lrs_solver == "analytical") {
        //Analytical solution to find roots
	t.start_now();
	min = solve_cubic(coeffs, lb);
	t.stop_now();
	t_solve_opt += t.user();
	TracePrint("Minimum computed using analytical method : %e\n", min);
      } else {
        //Numerical minimization using Newton-Raphson Method
	t.start_now();
	func f(coeffs);
	min = rtsafe(f, 0, sg_->get_var_ub(which_var), TOL);
	t.stop_now();
	t_solve_opt += t.user();
	TracePrint("Minimum computed using NR method : %e\n", min); 
      }
      if ( min < 0 )
	min = lb;
      break;
    }
  
#if 1
  int action = update_var(which_var, p, min, lb, ub);
#endif

#if 0
  int action = update_var(which_var, min, lb, ub);
#endif
  
  misc::cstr_t actions[] = {"no change", "updated", "set min", "set max"};
  
  TracePrint("update var %lu, %s, (lb : %e, ub : %e), modified ub : %e, order %lu minimization, min : %e, action : %s\n", 
	     which_var,
	     sg_->get_var_name(which_var),
	     lb,
	     sg_->get_var_ub(which_var),
	     ub,
	     p.get_order(),
	     min,
	     actions[action]);
  TracePrint("Time spent in cost function eval : %g, Optimization : %g\n", 
	     t_cost_eval, t_solve_opt);
  
  return action != NO_CHANGE;
}

double complex_cost_vars::
solve_quadratic(std::vector<double> coeffs)
{
  double a = coeffs[2];
  double b = coeffs[1];
  double c = coeffs[0];
  
  double D = b*b - 4*a*c;
  assert( D >= 0 && "Determinant of quadratic solution must be great than zero");
  double root1 = (-b + sqrt(D))/(2*a);
  //double root2 = (-b - sqrt(D))/(2*a);
  return root1;
}

//Solve a normalized cubic
double complex_cost_vars::
solve_cubic(std::vector<double> coeffs, double lb)
{
  double a = coeffs[3];
  double b = coeffs[2];
  double c = coeffs[1];
  double d = coeffs[0];
  
  double p = c/a - b*b/a/a/3;
  double q = (2*b*b*b/a/a/a - 9*b*c/a/a + 27*d/a) / 27;
  
  double pi = 3.141592654;
  
  double D = p*p*p/27 + q*q/4;
  double y1 = 0.0, y2 = 0.0, y3 = 0.0;
  
  double tfinal = b/a/3;
  
  //3 real unequal roots -- use trigonometric formulations
  if ( D < 0 ) {
    double phi = acos(-q/2/sqrt(abs(p*p*p)/27));
    double temp1=2*sqrt(abs(p)/3);
    y1 =  temp1*cos(phi/3);
    y2 = -temp1*cos((phi+pi)/3);
    y3 = -temp1*cos((phi-pi)/3);
    
    y1 -= tfinal;
    y2 -= tfinal;
    y3 -= tfinal;
    //printf("Roots y1 %e y2 %e y3 %e\n", y1, y2, y3);
  }
  else if ( D == 0 ){
    double temp = -q/2;
    double u = pow(abs(temp), 1.0/3.0);
    if ( temp < 0 ) u = -u;
    y1 = 2*u;
    y2 = -u;
    y3 = -u;
    
    y1 -= tfinal;
    y2 -= tfinal;
    y3 -= tfinal;
    //printf("Roots y1 %e y2 %e y3 %e\n", y1, y2, y3);
  } else {
    double temp1 = -q/2 + sqrt(D);
    double temp2 = -q/2 - sqrt(D);
    
    double u = pow(abs(temp1), 1.0/3.0);
    double v = pow(abs(temp2), 1.0/3.0);
    if ( temp1 < 0 ) u = -u;
    if ( temp2 < 0 ) v = -v;
    y1 = u + v;
    y2 = -1;
    y3 = -1;
    y1 -= tfinal;
    //printf("Roots y1 %e y2 %e y3 %e\n", y1, y2, y3);
  }
  
  std::vector<double> positive_roots;
  if ( y1 > 0 )
    positive_roots.push_back(y1);
  if ( y2 > 0 )
    positive_roots.push_back(y2);
  if ( y3 > 0 )
    positive_roots.push_back(y3);
  
  if ( positive_roots.size() > 1 ) {
    //now we check which roots lies between lb and ub
    std::sort(positive_roots.begin(), positive_roots.end());
    for(size_t i = 0; i < positive_roots.size(); i++) {
      if ( positive_roots[i] > lb )
	return positive_roots[i];
    }
    return positive_roots[positive_roots.size()-1];
  } else {
    return positive_roots[0];
  }
}

double complex_cost_vars::
generate_constraints_bound(size_t which_var)
{
  double ub = sg_->get_var_ub(which_var);
  
  double var_ub = ub;
  
  for(size_t i = 0; i < sg_->num_packing_constraints(); i++) {
    const math::posynomial &dp = sg_->get_constraint(i);
    double constraint_ub = sg_->get_constraint_ub(i);
    double temp_ub = constraint_ub - (dp.evaluate(get_vars()) - vars_[which_var]);
    if ( var_ub > temp_ub ) var_ub = temp_ub;
  }
  return var_ub;
}
 
void complex_cost_vars::
generate_polynomial(size_t which_var, complex_cost_vars::polynomial &p)
{
  TraceScope("GTSIZING", 10);
  
  //This is the linear term whose derivative will be a constant
  double coef_1 = sg_->get_var_weight(which_var);
  p.insert(1,coef_1);
  
  for (sizing_graph::id_crange cr = sg_->get_related_edges(which_var);
       cr.first != cr.second; ++cr.first)
    {
      size_t e = *cr.first;
      const math::posynomial &dp = sg_->get_edge_delay(e);
      std::string p_str = dp.to_string("x");
      for (math::posynomial::sum_term_crange ctr = dp.get_term_range();
	   ctr.first != ctr.second; ++ctr.first)
	{
	  std::pair<double, int> result
	    = ctr.first->partial_evaluate(get_vars(), which_var);
	  if ( mps_[e]*result.first != 0 )
	    p.insert(result.second, mps_[e]*result.first);
	}
    }
#if 1
  const math::posynomial cost = sg_->get_cost_function();
  for (math::posynomial::sum_term_crange ctr = cost.get_term_range();
       ctr.first != ctr.second; ++ctr.first)
    {
      std::pair<double, int> result
	= ctr.first->partial_evaluate(get_vars(), which_var);
      if ( result.first != 0 ) 
	p.insert(result.second, result.first);
    }
#endif
  polynomial::pc_map_iter iter = p.coeffs.begin();
  TracePrint("1-D optimization function for var %lu : %s \n", 
	     which_var,
	     sg_->get_var_name(which_var));
  for(; iter != p.coeffs.end(); iter++) {
    polynomial::pc_pair pcp = *iter;
    TracePrint("%e x^{%d} + ", pcp.second, pcp.first);
  }
  TracePrint("\n");
}

void complex_cost_vars::
generate_derivative(complex_cost_vars::polynomial p, std::vector<double>& coeffs)
{
  TraceScope("GTSIZING", 10);
  
  polynomial dp;
  polynomial::pc_map_iter iter = p.coeffs.begin();
  for(; iter != p.coeffs.end(); iter++) {
    polynomial::pc_pair pcp = *iter;
    int p = pcp.first;
    double c = p*pcp.second;
    if ( c != 0 ) dp.insert(p-1, c);
  }
  
  TracePrint("1-D derivative function\n");
  for(iter = dp.coeffs.begin(); iter != dp.coeffs.end(); iter++) {
    polynomial::pc_pair pcp = *iter;
    TracePrint("%e x^{%d} + ", pcp.second, pcp.first);
  }
  TracePrint("\n");
  
  coeffs.resize(dp.get_order()+1, 0.0);
  
  iter = dp.coeffs.begin();
  int min_power = dp.get_min_power();
  TracePrint("Minimum power of derivative : %d, Order : %lu\n", 
	     min_power, dp.get_order());
  
  for(; iter != dp.coeffs.end(); iter++) {
    polynomial::pc_pair pcp = *iter;
    int p = pcp.first;
    size_t index = p - min_power;
    coeffs[index] = pcp.second;
  }
  
  TracePrint("Coefficients for analytical computation\n");
  for(size_t i = 0; i < coeffs.size(); i++) {
    TracePrint("c[%lu] : %e, ", i, coeffs[i]);
  }
  TracePrint("\n");
  
  return;
}
