#include "../misc/common.h"
#include "../misc/trace.h"
#include "../misc/options.h"

#include "dual_sizing.h"

#define TOL 1e-6

RegisterOption(max_rounds, "-1", "maximum # of rounds for dual sizing")
RegisterOption(max_user, "-1", "maximum user seconds for dual sizing")
RegisterOption(min_change, "1e-3", "minimum change in cost/size to continue")
RegisterOption(min_period, "false", "minimize clock period, path only")
RegisterOption(weight_T, "0.01", "initial weight of T^2, path only")
RegisterOption(weight_r, "1.5", "expanding factor for weight of T^2, path only")
RegisterOption(max_paths, "10", "maximum # of path steps, path only")

using misc::options;

void dual_sizing::set_multipliers(double w)
{
	for (size_t i = 0; i < vars_->num_edges(); ++i)
		vars_->set_mp(i, flow_[i]+df_[i]*w);
}

void dual_sizing::get_mp_as_flow()
{
	for (size_t i = 0; i < vars_->num_edges(); ++i)
		// grab flow from mp to avoid accuracy issues
		flow_[i] = vars_->get_mp(i);
}

void dual_sizing::update_flow(double w)
{
	compute_LRS(w);

	get_mp_as_flow();
}

double dual_sizing::compute_derivative() const
{
	return std::inner_product(df_.begin(), df_.end(), vars_->get_delays(), 0.0);
}

static double compute_imp_ratio(double new_value, double old_value)
{
	double d = fabs(new_value-old_value);

	if (fabs(old_value) < TOL)
	{
		if (d < TOL)
			return 0;
		else
			return d/TOL;
	}
	else
	{
		return fabs(d/old_value);
	}
}

bool dual_sizing::should_finish(int round)
{
	double max_size_change = 0;
	for (size_t i = 0; i < vars_->num_vars(); ++i)
		max_size_change = std::max(max_size_change,
			compute_imp_ratio(vars_->get_var(i), old_vars_[i]));

	double max_flow_change = 0;
	double max_flow = 0;
	for (size_t i = 0; i < vars_->num_edges(); ++i)
	{
		max_flow = std::max(max_flow, flow_[i]);
		max_flow_change = std::max(max_flow_change,
			compute_imp_ratio(flow_[i], old_flows_[i]));
	}

	double cost_change = compute_imp_ratio(vars_->get_cost(), old_cost_);

	fprintf(stdout, "changes: size %.3f, flow %.3f, cost %e, no imp %lu\n",
		max_size_change, max_flow_change, cost_change, cost_imp_count_);

	if (feasible_found_)
		fprintf(stdout, "statics: flow %.3f, scale %.0f, gap ub %e\n",
			max_flow, scale_base_, feasible_target_-vars_->get_cost());
	else
		fprintf(stdout, "statics: flow %.3f, scale %.0f\n",
			max_flow, scale_base_);

	if (cost_change > min_change_)
	{
		cost_imp_count_ = 0;
		return false;
	}

	++cost_imp_count_;

	if (max_size_change > min_change_)
	{
		return false;
	}

	if (cost_imp_count_ >= 10)
	{
		fprintf(stdout, "  terminated when there is no change\n");
		return true;
	}

	if ((cost_imp_count_ >= 5) && (max_flow_change < min_change_))
	{
		fprintf(stdout, "  terminated when there is no change (and in flow)\n");
		return true;
	}

	return false;
}

bool dual_sizing::solve(sizing_vars *p_vars, size_t pre_solve_rounds)
{
	TraceScope("GTSIZING", 1);

	os::timer2 t;

	t.start_now();

	attach_vars(p_vars);

	int max_rounds = options::get().as_int("max_rounds");
	double max_user = options::get().as_double("max_user");

	if (pre_solve_rounds > 0)
	{
		max_user = -1;
		max_rounds = (int)pre_solve_rounds;
	}

	double desired_cp = vars_->get_sizing_graph()->get_clock_period();

	int round = 0;
	bool feasible = true;

	try
	{
	  compute_LRS(0);
	  
	  for (;; ++round)
	    {
	      t.stop_now();
	      
	      double cp = vars_->get_achieved_clock_period();
	      
	      fprintf(stdout,
		      "$%sround %d, cost %.2f, gap %.3f, cp %.2f, wns %.3f, "
		      "LRS %lu, LRS_iters %lu, user %.3f\n",
		      (pre_solve_rounds > 0)? "pp": "",
		      round, vars_->get_cost(), -vars_->get_fd(),
		      cp, desired_cp-cp,
		      LRS_.num_items(), LRS_iters_, t.user());
	      
	      TracePrint(
			 "  before LRS %e, check cost %e, target %e\n",
			 before_LRS_,
			 fabs(vars_->compute_cost()-vars_->get_cost()),
			 vars_->get_target());
	      
	      if (cp <= desired_cp) // primal feasible
		{
		  fprintf(stdout,
			  "primal feasible found: target %.2f\n",
			  vars_->get_target());
		  
		  if (!feasible_found_ || (feasible_target_ > vars_->get_target()))
		    {
		      feasible_found_ = true;
		      feasible_target_ = vars_->get_target();
		      feasible_cp_ = cp;
		      feasible_cost_ = vars_->get_cost();
		    }
		}
	      
	      //			if (cost_imp_count_ > 10)
	      //			{
	      //				if (cp <= desired_cp)
	      //					if ((feasible_target_-vars_->get_cost())/feasible_target_ < min_change_)
	      //						break;
	      //				if ((cp-desired_cp)/desired_cp < min_change_)
//					break;
//			}

			old_cost_ = vars_->get_cost();
			std::copy(vars_->get_vars(),
				vars_->get_vars()+vars_->num_vars(),
				old_vars_.begin());
			std::copy(flow_.begin(), flow_.end(), old_flows_.begin());

			if ((max_rounds > 0) && (round >= max_rounds))
				break;

			if ((max_user > 0) && (t.user() >= max_user))
				break;

			if (!find_next_flow(round))
			{
				++round;
				break;
			}

			if (should_finish(round))
			{
				++round;
				break;
			}
		}
	}
	catch (std::exception &e)
	{
		if (strcmp(e.what(), "not feasible") != 0)
			throw;

		get_mp_as_flow();

		++round;
		feasible = false;
	}

	t.stop_now();

	double cp = vars_->get_achieved_clock_period(true);

	if (pre_solve_rounds == 0)
	{
		fprintf(stdout, "@final %s\n", feasible? "feasible": "infeasible");

		fprintf(stdout, "@cost %.2f, target %.2f, gap %e, cost_ub %.3e\n",
			vars_->get_cost(), vars_->get_target(),
			-vars_->get_fd(), vars_->get_cost_ub());

		fprintf(stdout, "@cp %.2f, target_cp %.2f, wns %.3f\n",
			cp, desired_cp, desired_cp-cp);

		if (feasible_found_)
			fprintf(stdout, "@fcp %.2f, fcost %.2f, ftarget %.2f, gap_ub %e\n",
				feasible_cp_, feasible_cost_, feasible_target_,
				feasible_target_-vars_->get_cost());
		else
			fprintf(stdout, "@fcp unk, fcost unk, ftarget unk, gap_ub unk\n");

		TracePrint("  before LRS %e, check cost %e\n",
			before_LRS_, fabs(vars_->compute_cost()-vars_->get_cost()));

		fprintf(stdout, "@rounds %d, DF %lu, DF_user %.1f, "
			"LRS %lu, LRS_user %.1f, LRS_iters %lu\n",
			round, DF_.num_items(), DF_.total_user(),
			LRS_.num_items(), LRS_.total_user(), LRS_iters_);

		fprintf(stdout, "@user %.3f, sys %.3f, real %.3f\n",
			t.user(), t.sys(), t.real());
	}
	else
	{
		fprintf(stdout, "@ppfinal %s\n", feasible? "feasible": "infeasible");

		fprintf(stdout, "@ppcost %.2f, pptarget %.2f\n",
			vars_->get_cost(), vars_->get_target());

		fprintf(stdout, "@ppcp %.2f, ppwns %.3f\n",
			cp, desired_cp-cp);

		if (feasible_found_)
			fprintf(stdout, "@ppfcp %.2f, ppfcost %.2f, ppftarget %.2f\n",
				feasible_cp_, feasible_cost_, feasible_target_);

		TracePrint("  before LRS %e, check cost %e\n",
			before_LRS_, fabs(vars_->compute_cost()-vars_->get_cost()));

		fprintf(stdout, "@pprounds %d, ppDF %lu, ppDF_user %.1f, "
			"ppLRS %lu, ppLRS_user %.1f, ppLRS_iters %lu\n",
			round, DF_.num_items(), DF_.total_user(),
			LRS_.num_items(), LRS_.total_user(), LRS_iters_);

		fprintf(stdout, "@ppuser %.3f, ppsys %.3f, ppreal %.3f\n",
			t.user(), t.sys(), t.real());
	}

	return feasible;
}

namespace {

std::string not_feasible("not feasible");

} // namespace

size_t dual_sizing::compute_LRS(double w, bool check_feasible)
{
	LRS_.start_now();

	set_multipliers(w);

	before_LRS_ = vars_->get_cost();

	size_t iters = vars_->solve_LRS();

	LRS_.stop_now();

	LRS_iters_ += iters;

	if (check_feasible)
		if (vars_->get_cost() > vars_->get_cost_ub())
			throw std::runtime_error(not_feasible);

	return iters;
}

double dual_sizing::compute_cost() const
{
	return vars_->compute_cost();
}

dual_sizing::dual_sizing()
	: vars_(0),
	before_LRS_(0),
	old_cost_(0),
	min_change_(options::get().as_double("min_change")),
	LRS_iters_(0),
	cost_imp_count_(0),
	scale_base_(10000),
	feasible_found_(false),
	feasible_cp_(0),
	feasible_cost_(0),
	feasible_target_(0)
{
}

void dual_sizing::attach_vars(sizing_vars *p_vars)
{
	vars_ = p_vars;

	const size_t N = vars_->num_vertices();
	const size_t M = vars_->num_edges();

	// set initial mps using vars
	// this enables passing result from one method to another
	flow_.assign(vars_->get_mps(), vars_->get_mps()+M);

	df_.resize(M);
	dfi_.resize(M);

	old_vars_.resize(vars_->num_vars());
	old_flows_.resize(M);

	cs_solver_.assign(N, vars_->get_sizing_graph()->get_edges(), M);
}

double dual_sizing::update_direction(double s)
{
	TraceScope("GTSIZING", 3);

	DF_.start_now();

	double md = 0;
	for (size_t i = 0; i < vars_->num_edges(); ++i)
		md = std::max(md, fabs(vars_->get_delay(i)));

	double dscale = 100; //scale_base_/md;

	double fscale = scale_base_/s;

	TracePrint("  max_delay %.3e, delay_scale %.3e, flow_step %.3e, flow_scale %.3e\n",
		md, dscale, s, fscale);

	for (size_t i = 0; i < vars_->num_edges(); ++i)
	{
		long ub = long(s*fscale);
		long lb = (flow_[i] < s)? -long(flow_[i]*fscale): -ub;
		long long cost = -(long long)(vars_->get_delay(i)*dscale);

		cs_solver_.set_edge_caps(i, lb, ub);
		cs_solver_.set_edge_cost(i, cost);
		cs_solver_.set_edge_flow(i, 0);
	}

//	bool succ = cs_solver_.solve_unknown();
	bool succ = cs_solver_.solve_feasible();
	assert(succ);

	cs_solver_.collect_edge_flows(&dfi_[0]);

	for (size_t i = 0; i < vars_->num_edges(); ++i)
		df_[i] = dfi_[i]/fscale;

	DF_.stop_now();

	return compute_derivative();
}

void dual_sizing::solve_path(sizing_vars *p_vars)
{
	TraceScope("GTSIZING", 1);

	os::timer2 t;

	t.start_now();

	attach_vars(p_vars);

	int max_rounds = options::get().as_int("max_rounds");
	int max_paths = options::get().as_int("max_paths");
	double max_user = options::get().as_double("max_user");
	bool min_period = options::get().as_bool("min_period");
	double weight_T = options::get().as_double("weight_T");
	double weight_r = options::get().as_double("weight_r");

	double desired_cp = vars_->get_sizing_graph()->get_clock_period();

	compute_LRS(0, false); // let sizes settled for initial mps

	// bound
	double cp_lb = min_period? 0: desired_cp;
	double cp_ub = 1e10;

	int round = 0;
	int path = 0;
	for (;; ++path, weight_T *= weight_r)
	{
		t.stop_now();

		vars_->set_T_2(weight_T, cp_lb, cp_ub);

		double cp = vars_->get_achieved_clock_period();
		double T = vars_->get_T();
		double dual = vars_->get_T_2_dual_cost();
		double primal = vars_->get_T_2_primal_cost();
		double origin = vars_->get_T_2_original_cost();

		fprintf(stdout,
			"$path %d, weight_T %.3f, "
			"dual %.2f, primal %.2f, origin %.2f, "
			"cp %.2f, T %.2f, user %.3f\n",
			path, weight_T, dual, primal, origin,
			cp, T, t.user());

		if ((max_paths > 0) && (path >= max_paths))
			break;

		if ((max_user > 0) && (t.user() >= max_user))
			break;

		if ((max_rounds > 0) && (round >= max_rounds))
			break;

		for (;; ++round)
		{
			t.stop_now();

			double cp = vars_->get_achieved_clock_period();
			double T = vars_->get_T();
			double dual = vars_->get_T_2_dual_cost();
			double primal = vars_->get_T_2_primal_cost();
			double origin = vars_->get_T_2_original_cost();

			fprintf(stdout,
				"$round %d, path %d, dual %.2f, primal %.2f, origin %.2f, "
				"cp %.2f, T %.2f, LRS %lu, LRS_iters %lu, user %.3f\n",
				round, path, dual, primal, origin, cp, T,
				LRS_.num_items(), LRS_iters_, t.user());

			if (min_period)
			{
				if (!feasible_found_ || (cp < feasible_cp_))
				{
					fprintf(stdout,
						"primal min cp found: cp %.2f, target %.2f\n",
						cp, origin);

					feasible_found_ = true;
					feasible_target_ = origin;
					feasible_cp_ = cp;
				}
			}
			else
			{
				if (cp <= desired_cp) // primal feasible
				{
					fprintf(stdout,
						"primal feasible found: target %.2f\n",
						origin);

					if (!feasible_found_ || (feasible_target_ > origin))
					{
						feasible_found_ = true;
						feasible_target_ = origin;
						feasible_cp_ = cp;
					}
				}
			}

			if ((max_rounds > 0) && (round >= max_rounds))
				break;

			if ((max_user > 0) && (t.user() >= max_user))
				break;

			if (!find_next_flow(round))
			{
				++round;
				break;
			}

			if (fabs((primal-dual)/primal) < sqrt(min_change_))
			{
				++round;
				break;
			}
		}
	}

	t.stop_now();

	fprintf(stdout, "@final feasible\n");

	double cp = vars_->get_achieved_clock_period();
	double dual = vars_->get_T_2_dual_cost();
	double primal = vars_->get_T_2_primal_cost();
	double origin = vars_->get_T_2_original_cost();

	fprintf(stdout, "@weight_T %.3f, dual %.2f, primal %.2f, origin %.2f\n",
		weight_T, dual, primal, origin);

	fprintf(stdout, "@cp %.2f, target_cp %.2f, wns %.3f\n",
		cp, desired_cp, cp-desired_cp);

	if (feasible_found_)
		fprintf(stdout, "@fcp %.2f, ftarget %.2f\n",
			feasible_cp_, feasible_target_);
	else
		fprintf(stdout, "@fcp unk, ftarget unk\n");

	fprintf(stdout, "@paths %d, rounds %d, DF %lu, DF_user %.1f, "
		"LRS %lu, LRS_user %.1f, LRS_iters %lu\n",
		path, round, DF_.num_items(), DF_.total_user(),
		LRS_.num_items(), LRS_.total_user(), LRS_iters_);

	fprintf(stdout, "@user %.3f, sys %.3f, real %.3f\n",
		t.user(), t.sys(), t.real());
}
