#include "../misc/common.h"


#include "../misc/trace.h"


#include "../misc/timer.h"


#include "../misc/options.h"





#include "flow_sizing.h"


#include "flow_solver.h"





#define TOL 1e-6





RegisterTraceScope(FLOW, 1, "show dual maximizing details")





RegisterOption(line_epsi, "0.02", "expected increase param for inexact line search")


RegisterOption(line_alpha, "2", "step size param for inexact line search")


RegisterOption(cg_dir, "false", "use conjugate direction")


RegisterOption(dual_rounds, "100", "maximum # of dual rounds")





void flow_sizing::collect_delays_and_bounds(


	int D_scale,


	int f_scale,


	double f_min_pos)


{


	for (size_t i = 0; i < vars_->num_edges(); ++i)


	{


		if (flow_[i] < 10)


			f_bound_[2*i] = -int(f_scale*flow_[i]/10);


		else


			f_bound_[2*i] = -f_scale;





		f_bound_[2*i+1] = f_scale;





		D_[i] = int(D_scale*vars_->get_delay(i));


    }


}





void flow_sizing::set_multipliers(double w)


{


	for (size_t i = 0; i < vars_->num_edges(); ++i)


		vars_->set_mp(i, flow_[i]+df_[i]*w);


}





void flow_sizing::update_flow(double w)


{


	for (size_t i = 0; i < vars_->num_edges(); ++i)


		flow_[i] += df_[i]*w;


}





double flow_sizing::compute_derivative() const


{


	return std::inner_product(df_.begin(), df_.end(), vars_->get_delays(), 0.0);


}





double flow_sizing::compute_max_w() const


{


	double ub = 0;


	bool has_max = false;





	for (size_t i = 0; i < vars_->num_edges(); ++i)


	{


		if (df_[i] >= 0)


			continue;





		double w = -flow_[i]/df_[i];





		if (!has_max || (w < ub))


			ub = w;





		has_max = true;


	}





	if (has_max)


		return ub;


	else


		return -1;


}





bool flow_sizing::solve_dual(sizing_vars *p_vars, int max_iter, bool exact)


{


	epsi_ = misc::options::get().as_double("line_epsi");


	alpha_ = misc::options::get().as_double("line_alpha");


	cg_ = misc::options::get().as_bool("cg_dir");


	cg_count_ = 0;





	TraceScope("FLOW", 1);





	attach_vars(p_vars);





	feasible_ = true;





	int i = 0;





	os::timer_t t;





	for (; i < max_iter; ++i)


	{


		size_t iters = compute_LRS(0);





		t.stop();


		fprintf(stdout,


			"dual round %d, cost %e, target %e, iters %u, "


			"total_iters %u, LRS %u, duality gap %e, real %.3f\n",


			i, vars_->get_cost(), vars_->get_target(),


			iters, total_iters_, LRS_count_, vars_->get_fd(), t.real());





		TracePrint("    before LRS %e, check cost %e\n",


			before_LRS_, fabs(vars_->compute_cost()-vars_->get_cost()));





		if (!feasible_)


			break;





		if (!(exact? find_next_flow(): find_next_flow_inexact()))


			break;


	}





	size_t iters = compute_LRS(0);





	if (feasible_)


	{


		fprintf(stdout,


			"final feasible, rounds %d, cost %e, target %e, "


			"iters %u, total_iters %u, LRS %u\n",


			i, vars_->get_cost(), vars_->get_target(),


			iters, total_iters_, LRS_count_);


	}


	else


	{


		fprintf(stdout,


			"final infeasible, rounds %d, cost %e, target %e, "


			"cost_ub %e, iters %u, total_iters %u, LRS %u\n",


			i, vars_->get_cost(), vars_->get_target(),


			vars_->get_cost_ub(), iters, total_iters_, LRS_count_);


	}





	TracePrint("    before LRS %e, check cost %e\n",


		before_LRS_, fabs(vars_->compute_cost()-vars_->get_cost()));


	


	FILE * fpt = fopen("var_flow_sizing.dump","w");


	vars_->dump(fpt);


	fclose(fpt);





	return feasible_;


}





bool flow_sizing::solve_dual(sizing_vars *p_vars, bool exact)


{


	epsi_ = misc::options::get().as_double("line_epsi");


	alpha_ = misc::options::get().as_double("line_alpha");


	cg_ = misc::options::get().as_bool("cg_dir");


	cg_count_ = 0;





	TraceScope("FLOW", 1);





	attach_vars(p_vars);





	feasible_ = true;





	int max_iter = misc::options::get().as_int("dual_rounds");


	int i = 0;





	os::timer_t t;





	for (; i < max_iter; ++i)


	{


		size_t iters = compute_LRS(0);





		t.stop();


		fprintf(stdout,


			"dual round %d, cost %e, target %e, Achieved delay %e, iters %u, "


			"total_iters %u, LRS %u, duality gap %e, real %.3f\n",


			i, vars_->get_cost(), vars_->get_target(), vars_->get_achieved_delay(),


			iters, total_iters_, LRS_count_, vars_->get_fd(), t.real());





		TracePrint("    before LRS %e, check cost %e\n",


			before_LRS_, fabs(vars_->compute_cost()-vars_->get_cost()));





		if (!feasible_)


			break;





		if (!(exact? find_next_flow(): find_next_flow_inexact()))


			break;


	}





	size_t iters = compute_LRS(0);





	if (feasible_)


	{


		fprintf(stdout,


			"final feasible, rounds %d, cost %e, target %e, Achieved delay %e,"


			"iters %u, total_iters %u, LRS %u\n",


			i, vars_->get_cost(), vars_->get_target(),vars_->get_achieved_delay(),


			iters, total_iters_, LRS_count_);


	}


	else


	{


		fprintf(stdout,


			"final infeasible, rounds %d, cost %e, target %e, Achieved delay %e"


			"cost_ub %e, iters %u, total_iters %u, LRS %u\n",


			i, vars_->get_cost(), vars_->get_target(), vars_->get_achieved_delay(),


			vars_->get_cost_ub(), iters, total_iters_, LRS_count_);


	}





	TracePrint("    before LRS %e, check cost %e\n",


		before_LRS_, fabs(vars_->compute_cost()-vars_->get_cost()));


	


	FILE * fpt = fopen("var_flow_sizing.dump","w");


	vars_->dump(fpt);


	fclose(fpt);





	return feasible_;


}





size_t flow_sizing::compute_LRS(double w)


{


	fflush(stdout);





	set_multipliers(w);





	before_LRS_ = vars_->get_cost();





	size_t iters = vars_->solve_LRS();





	total_iters_ += iters;





	feasible_ = (vars_->get_cost() <= vars_->get_cost_ub());





	++LRS_count_;





	return iters;


}





bool flow_sizing::find_next_flow()


{


	TraceScope("FLOW", 2);





	if (!update_direction())


		return false;





	double wl = 0;


	double drl = compute_derivative();


	double costl = vars_->get_cost();





	TracePrint("linear round 0, w = 0, g' = %+.1e, g = %.3e (%.3e)\n",


		drl, costl, before_LRS_);


	


	if (drl < TOL)


		return false;


	


	double wr = compute_max_w();


	double drr = 0;


	double costr = 0;





	if (wr < 0) // not bounded by feasibility of flow


	{


		size_t i = 0;


		for (wr = 1;; wr *= 2, ++i)


		{


			size_t iters = compute_LRS(wr);


			drr = compute_derivative();


			costr = vars_->get_cost();





			TracePrint("max_w round %u, w = %.3e, g' = %+.1e, "


				"g = %.3e (%.3e), iters %u, total_iters %u, LRS %u\n",


				i, wr, drr, costr, before_LRS_, iters, total_iters_,


				LRS_count_);





			if (!feasible_)


			{


				update_flow(wr);


				return false;


			}





			if (drr < TOL)


				break;





			wl = wr, drl = drr, costl = costr;


		}


	}


	else


	{


		size_t iters = compute_LRS(wr);


		drr = compute_derivative();


		costr = vars_->get_cost();





		TracePrint("max_w probe, w = %.3e, g' = %+.1e, "


			"g = %.3e (%.3e), iters %u, total_iters %u, LRS %u\n",


			wr, drr, costr, before_LRS_, iters, total_iters_, LRS_count_);





		if (!feasible_)


		{


			update_flow(wr);


			return false;


		}


	}





	if (drr < TOL) // max_w is decreasing


	{


		for (int i = 1;; ++i)


		{


			double wm = (wl+wr)/2;


			//double wm = (wr*drl-wl*drr)/(drl-drr);


	


			size_t iters = compute_LRS(wm);


	


			double drm = compute_derivative();


			double costm = vars_->get_cost();





			TracePrint(


				"linear round %d, w = %.3e, g' = %+.1e, g = %.3e (%.3e), "


				"iters %u, total_iters %u, LRS %u, range %.1e/%.1e\n",


				i, wm, drm, costm, before_LRS_, iters, total_iters_,


				LRS_count_, wr-wl, fabs(costr-costl));





			if (!feasible_)


			{


				update_flow(wm);


				return false;


			}





			if (drm >= 0)


				wl = wm, drl = drm, costl = costm;


			else


				wr = wm, drr = drm, costr = costm;





			if ((fabs(drr) < TOL)


				|| (fabs(drl) < TOL)


				|| (wr-wl < TOL*sqrt(TOL))


				|| (fabs(costr-costl) < TOL))


				break;


		}


	}





	if (costr > costl)


	{


		update_flow(wr);


		return wr > TOL*sqrt(TOL);


	}


	else


	{


		update_flow(wl);


		return wl > TOL*sqrt(TOL);


	}


}





double flow_sizing::compute_cost() const


{


	return vars_->compute_cost();


}





flow_sizing::flow_sizing()


	: vars_(0),


	step_(1),


	epsi_(0.02),


	alpha_(2),


	cg_(false),


	before_LRS_(0),


	total_iters_(0),


	LRS_count_(0),


	cs_count_(0),


	feasible_(true)


{


}





void flow_sizing::attach_vars(sizing_vars *p_vars)


{


	vars_ = p_vars;





	const size_t M = vars_->num_edges();


	flow_.assign(M, 0);





	D_.resize(M);


	f_bound_.resize(2*M);


	df_.resize(M);


	dfi_.resize(M);


	df_last_.resize(M);


}





bool flow_sizing::find_next_flow_inexact()


{


	TraceScope("FLOW", 2);





	if (!update_direction())


		return false;





	double dr0 = compute_derivative();


	double cost0 = vars_->get_cost();





	TracePrint("inexact start, g' = %+.1e, g = %.3e (%.3e)\n",


		dr0, cost0, before_LRS_);


	


	if (dr0 < TOL)


		return false;


	


	double w_max = compute_max_w();





	if ((w_max >= 0) && (step_ > w_max))


		step_ = w_max;





	size_t iters = compute_LRS(step_);


	double dr = compute_derivative();


	double cost = vars_->get_cost();





	TracePrint("inexact round 0, w = %.3e, g' = %+.1e, "


		"g = %.3e (%.3e), iters %u, total_iters %u, LRS %u\n",


		step_, dr, cost, before_LRS_, iters, total_iters_, LRS_count_);





	if (!feasible_)


	{


		update_flow(step_);


		return false;


	}





	size_t i = 1;


	if (cost > cost0+dr0*step_*epsi_) // not too big, increase step


	{


		for (step_ *= alpha_; (w_max < 0) || (step_ <= w_max); step_ *= alpha_, ++i)


		{


			iters = compute_LRS(step_);


			dr = compute_derivative();


			cost = vars_->get_cost();





			TracePrint("inexact round %u, w = %.3e, g' = %+.1e, "


				"g = %.3e (%.3e), iters %u, total_iters %u, LRS %u\n",


				i, step_, dr, cost, before_LRS_, iters, total_iters_,


				LRS_count_);





			if (!feasible_)


			{


				update_flow(step_);


				return false;


			}





			if (cost <= cost0+dr0*step_*epsi_) // big enough


				break;


		}


		step_ /= alpha_;


	}


	else // not too small, decrease step


	{


		for (step_ /= alpha_; step_ > TOL*sqrt(TOL); step_ /= alpha_, ++i)


		{


			iters = compute_LRS(step_);


			dr = compute_derivative();


			cost = vars_->get_cost();





			TracePrint("inexact round %u, w = %.3e, g' = %+.1e, "


				"g = %.3e (%.3e), iters %u, total_iters %u, LRS %u\n",


				i, step_, dr, cost, before_LRS_, iters, total_iters_,


				LRS_count_);





			if (!feasible_)


			{


				update_flow(step_);


				return false;


			}





			if (cost > cost0+dr0*step_*epsi_) // small enough


				break;


		}


	}





	update_flow(step_);





	return step_ > TOL*sqrt(TOL);


}





bool flow_sizing::update_direction()


{


	TraceScope("FLOW", 7);





	collect_delays_and_bounds(10000, 10000, 0.1);





	char cs_name[1000];


	if (TraceActive())


		sprintf(cs_name, "%04d", cs_count_);


	else


		sprintf(cs_name, "last");


	++cs_count_;





	if (!compute_direction(


		vars_->get_sizing_graph()->num_vertices(),


		vars_->get_sizing_graph()->num_edges(),


		vars_->get_sizing_graph()->get_edges(),


		&D_[0],


		&f_bound_[0],


		&dfi_[0],


		cs_name))


		return false;





	if (cg_)


		std::copy(df_.begin(), df_.end(), df_last_.begin());





	std::copy(dfi_.begin(), dfi_.end(), df_.begin());





	if (cg_)


	{


		if (cg_count_ != 0)


		{


			--cg_count_;





			double mlast = 0, mcur = 0;


			for (size_t i = 0; i < df_.size(); ++i)


			{


				mlast += df_last_[i]*df_last_[i];


				mcur += df_[i]*df_[i];


			}


			double alpha = sqrt(mcur/mlast);


			for (size_t i = 0; i < df_.size(); ++i)


				df_[i] += df_last_[i]*alpha;


		}


		else


		{


			cg_count_ = vars_->num_vars();


		}


	}





	return true;


}


