#include "../misc/common.h"
#include "../misc/options.h"
#include "../misc/auto_file.h"
#include "../misc/trace.h"
#include "../misc/token_parser.h"

#include "flow_sizing.h"
#include "lr_sizing.h"
#include "flow_solver.h"
#include "posynomial.h"
#include "pmatrix.h"

#if 0
#include "complex_cost_vars_slow.h"
#endif 

#if 1
#include "complex_cost_vars.h"
#endif

using namespace math;

using misc::cstr_t;
using misc::options;

sizing_graph g_sg;
sizing_vars g_vars;
complex_cost_vars g_cost_vars;

//sizing_vars_elmore g_vars_elmore;

bool load_sizing_graph(cstr_t sg_name);
bool load_cost(cstr_t sg_name);

void do_optimize(sizing_vars *p_vars);

RegisterOption(solver, "flow", "solvers: flow|lr|fast_lr|tr|path|cplex")
RegisterOption(temperature, "true", "Temperature aware optimization") 

int usage(misc::cstr_t exefile)
{
	fprintf(stdout,
		"usage: %s sizing_graph options\n",
		exefile);

	options::get().show_options(stdout);

	return 1;					
}

void test_posynomial()
{
  std::string pm_file = "test.pm";
  FILE * fpt = fopen(pm_file.c_str(), "w");
  
  printf("Checking the API for dumping out the posynomial as string\n");
  
  posynomial p;
  posynomial::sum_term &st = p.add_term(2);
  st.add_term(0,1);
  st.add_term(1,2);
  posynomial p2;
  p2 = p+p;
  p2.save(fpt);
  printf("%s\n",p2.to_string("x",".3f").c_str());
  
  posynomial p1 = p*p;
  p1.save(fpt);
  printf("%s\n",p1.to_string("x",".3f").c_str());
  
  p1 = p1*p2;
  p1.save(fpt);
  printf("%s\n",p1.to_string("x",".3f").c_str());
  
  p2 = p1 + p2;
  p2.save(fpt);
  printf("%s\n",p2.to_string("x",".3f").c_str());
  
  p2 = p*p;
  p2 = p1 + p2;
  p2.save(fpt);
  printf("%s\n",p2.to_string("x",".3f").c_str());
  
  //Checking constant operation
  posynomial p3;
  p3.add_term(2);
  p2 = p*p3;
  p2 = p3 + p2;
  printf("%s\n",p2.to_string("x",".3f").c_str());
  
  //Checking constant 0 operation
  posynomial p4;
  p4.add_term(0);
  p2 = p2 + p4;
  printf("%s\n",p2.to_string("x",".3f").c_str());
  p2 = p2*p4;
  printf("%s\n",p2.to_string("x",".3f").c_str());
  //Checking matrix operation
  
  pmatrix pm(2,2,p);
  pm.save(fpt,"x",".3f");
  pmatrix pm2 = pm*pm;
  pm2 = pm + pm2;
  pm2.save(fpt,"x",".3f");
}

std::string get_benchmark_name(misc::cstr_t cmdline, size_t& length)
{
  std::vector<std::string> tokens, tokens1;
  cstr_t delimit = "/";
  misc::split_tokens(cmdline,&tokens, delimit);
  length = tokens[tokens.size()-1].length();
  
  delimit = ".";
  misc::split_tokens(tokens[tokens.size()-1].c_str(), &tokens1, delimit);
  
  return tokens1[0];
}

std::string get_path(cstr_t cmdline, size_t length)
{
  std::string cmd(cmdline);
  return cmd.substr(0,cmd.length() - length);
}

int main(int argc, char *argv[])
{
  options::get().load_options(argc, argv);
  misc::get_tracer().set_active_levels_using_options();

#if 0
  test_posynomial();
#endif
  
  if (argc < 2)
    return usage(argv[0]);
  
  options::get().show_options(stdout);
  
  if (!load_sizing_graph(argv[1]))
    return -1;
  
  size_t length = 0;
  std::string benchmark = get_benchmark_name(argv[1] , length);
  std::string path = get_path(argv[1], length);
  std::string cost_file = path + benchmark + ".cost";
  
  if ( !load_cost(cost_file.c_str()) )
    return -1;
  
  sizing_vars *p_vars;
  
  
  if (options::get().as_bool("temperature"))
    p_vars = &g_cost_vars;
  else
    p_vars = &g_vars;
  
  p_vars->attach_graph(&g_sg);
  
  do_optimize(p_vars);

  return 0;
}

bool load_cost(cstr_t sg_name)
{
  fprintf(stdout, "Cost file name : %s\n", sg_name);
  
  if ( !g_sg.load_cost(sg_name) )
    return false;
  
  fprintf(stdout, "@bench %s, cost function terms : %lu, packing constraints : %lu\n", g_sg.get_graph_name().c_str(),
	  g_sg.num_cost_terms(), g_sg.num_packing_constraints() );
  fflush(stdout);
  return true;
}

bool load_sizing_graph(cstr_t sg_name)
{
	if (!g_sg.load(sg_name))
	  return false;

	fprintf(stdout, "@bench %s, scale %.3e\n"
		"@vertices %lu, edges %lu, vars %lu, T %.3f\n",
		g_sg.get_graph_name().c_str(), g_sg.get_scale(),
		g_sg.num_vertices(), g_sg.num_edges(), g_sg.num_vars(),
		g_sg.get_clock_period());

	fflush(stdout);

	return true;
}

void do_optimize(sizing_vars *p_vars)
{
  if(options::get().as_str("solver") == std::string("lr"))
    {
      lr_sizing lr;
      
      lr.solve(p_vars);
    }
  else if(options::get().as_str("solver") == std::string("flow"))
    {
      flow_sizing fs;
      
      fs.solve(p_vars);
    }
  else if(options::get().as_str("solver") == std::string("cplex"))
    {
      fprintf(stdout, "We are doing a comparison with CPLEX runtime\n");
    }
  else if(options::get().as_str("solver") == std::string("fast_lr"))
    {
      flow_sizing fs;
      
      fs.solve(p_vars, 10);
      
      lr_sizing lr;
      
      lr.solve(p_vars);
    }
  else if(options::get().as_str("solver") == std::string("tr"))
    {
      flow_sizing_tr tr;
      
      tr.solve(p_vars);
    }
  else if(options::get().as_str("solver") == std::string("path"))
    {
      flow_sizing fs;
      
      fs.solve_path(p_vars);
    }
  else
    {
      fprintf(stderr, "error: unknown solver %s\n",
	      options::get().as_str("solver"));
    }
}
