#include "../misc/common.h"
#include "../misc/auto_file.h"
#include "../misc/options.h"
#include "../misc/counting_sort.h"

#include "flow_solver.h"

// grab types and functions from cs2

extern "C" {

#include "types_cs2.h"

void cs2_solve_unknown(long n, long m, node *nodes, arc *arcs, long f_sc, price_t max_c);
void cs2_solve_feasible(long n, long m, node *nodes, arc *arcs, long f_sc, price_t max_c);

void compute_prices();

}

namespace flow
{

CS_solver::~CS_solver()
{
}

CS_solver::CS_solver()
	: cost_(0)
{
}

CS_solver::CS_solver(size_t num_V, const size_t edges[], size_t num_E)
	: cost_(0)
{
	assign(num_V, edges, num_E);
}

CS_solver::node_t CS_solver::default_node() const
{
	node_t n;

	// this is a C struct, just use memset
	memset(&n, 0, sizeof(n));

	return n;
}

CS_solver::arc_t CS_solver::default_arc() const
{
	arc_t a;

	// this is a C struct, just use memset
	memset(&a, 0, sizeof(a));

	return a;
}

void CS_solver::init_arc_pair(size_t i_edge, size_t from, size_t to)
{
	assert(from < num_vertices());
	assert(to < num_vertices());

	arcs_[i_edge*2].head = &nodes_[to];
	arcs_[i_edge*2].sister = &arcs_[i_edge*2+1];
	arcs_[i_edge*2].arc_id = int(i_edge)+1;

	arcs_[i_edge*2+1].head = &nodes_[from];
	arcs_[i_edge*2+1].sister = &arcs_[i_edge*2];
	arcs_[i_edge*2+1].arc_id = -int(i_edge)-1;
}

struct arc_tail // : public std::unary_function<CS_solver::arc_t, size_t>
{
	arc_tail(CS_solver::node_t *p) : base_(p) {}

	size_t operator() (const CS_solver::arc_t &a) const
	{
		return a.sister->head-base_;
	}

protected:

	CS_solver::node_t *base_;

}; // struct arc_tail

struct arc_swap
{
	void operator() (CS_solver::arc_t &a, CS_solver::arc_t &b) const
	{
		assert(&a != &b);
		
		std::swap(a, b);

		if (a.sister == &a)
		{
			assert(b.sister == &b);

			a.sister = &b;
			b.sister = &a;
		}
		else
		{
			assert(b.sister != &b);
			assert(a.sister->sister == &b);
			assert(b.sister->sister == &a);

			a.sister->sister = &a;
			b.sister->sister = &b;
		}
	}

}; // struct arc_swap

void CS_solver::assign(size_t num_V, const size_t edges[], size_t num_E)
{
	nodes_.assign(num_V+1, default_node());
	
	arcs_.assign(2*num_E+1, default_arc());

	vertex_infos_.assign(num_V, vertex_info());
	edge_infos_.assign(num_E, edge_info());

	for (size_t i = 0; i < num_E; ++i)
		init_arc_pair(i, edges[i*2], edges[i*2+1]);

	std::vector<size_t> offsets;
	misc::counting_sort(
		arcs_.begin(),
		arcs_.begin()+2*num_E,
		num_V,
		&offsets,
		arc_tail(&nodes_[0]),
		arc_swap());

	assert(nodes_.size() == offsets.size());
	for (size_t i = 0; i < nodes_.size(); ++i)
		nodes_[i].suspended = &arcs_[offsets[i]];
}

size_t CS_solver::num_arcs() const
{
	return arcs_.size()-1;
}

size_t CS_solver::num_edges() const
{
	return num_arcs()/2;
}

size_t CS_solver::num_nodes() const
{
	return nodes_.size()-1;
}

size_t CS_solver::num_vertices() const
{
	return num_nodes();
}

bool CS_solver::backward_arc(size_t i_arc) const
{
	assert(i_arc < num_arcs());
	assert(arcs_[i_arc].arc_id != 0);

	return arcs_[i_arc].arc_id < 0;
}

bool CS_solver::forward_arc(size_t i_arc) const
{
	assert(i_arc < num_arcs());
	assert(arcs_[i_arc].arc_id != 0);

	return arcs_[i_arc].arc_id > 0;
}

void CS_solver::set_vertex_excess(size_t which_vertex, long long excess)
{
	assert(which_vertex < num_vertices());

	vertex_infos_[which_vertex].excess = excess;
}

void CS_solver::set_edge_cost(size_t which_edge, long long cost)
{
	assert(which_edge < num_edges());

	edge_infos_[which_edge].cost = cost;
}

void CS_solver::set_edge_caps(size_t which_edge, long lb, long ub)
{
	assert(which_edge < num_edges());
	assert(lb <= ub);

	edge_infos_[which_edge].lb = lb;
	edge_infos_[which_edge].ub = ub;
	edge_infos_[which_edge].flow = lb;
}

void CS_solver::set_edge_flow(size_t which_edge, long flow)
{
	assert(which_edge < num_edges());
	assert(edge_infos_[which_edge].lb <= flow);
	assert(flow <= edge_infos_[which_edge].ub);

	edge_infos_[which_edge].flow = flow;
}

long long CS_solver::get_vertex_excess(size_t which_vertex) const
{
	assert(which_vertex < num_vertices());

	return vertex_infos_[which_vertex].excess;
}

long long CS_solver::get_edge_cost(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return edge_infos_[which_edge].cost;
}

long CS_solver::get_edge_lb(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return edge_infos_[which_edge].lb;
}

long CS_solver::get_edge_ub(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return edge_infos_[which_edge].ub;
}

long CS_solver::get_edge_flow(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return edge_infos_[which_edge].flow;
}

long long CS_solver::get_cost() const
{
	return cost_;
}

CS_solver::vertex_info::vertex_info(long long exc)
	: excess(exc)
{
}

CS_solver::edge_info::edge_info(long l, long u, long f, long long c)
	: cost(c), lb(l), ub(u), flow(f)
{
}

long long CS_solver::initialize()
{
	long long total_excess = 0;

	for (size_t i = 0; i < num_nodes(); ++i)
	{
		long long exc = get_vertex_excess(i);
		nodes_[i].excess =exc ;
		total_excess += exc;
	}

	if (total_excess != 0)
		return -1;

	long long max_cost = 0;

	for (size_t i = 0; i < num_arcs(); ++i)
	{
		if (backward_arc(i))
			continue;

		arc_t *this_arc = &arcs_[i];
		arc_t *sis_arc = this_arc->sister;

		size_t i_edge = this_arc->arc_id-1;

		long lb = get_edge_lb(i_edge);
		long ub = get_edge_ub(i_edge);
		long flow = get_edge_flow(i_edge);
		long long cost = get_edge_cost(i_edge);

		this_arc->r_cap = ub-flow;
		sis_arc->r_cap = flow-lb;
		this_arc->head->excess += flow;
		sis_arc->head->excess -= flow;

		this_arc->cost = cost;
		sis_arc->cost = -cost;

		if (cost < 0)
			cost = -cost;

		max_cost = std::max(max_cost, cost);
	}

	return max_cost;
}

bool CS_solver::solve_unknown()
{
	long long max_c = initialize();

	if (max_c < 0)
		return false;

	cs2_solve_unknown(
		long(num_nodes()), long(num_arcs()),
		&nodes_[0], &arcs_[0],
		12, max_c);

	collect_everything();

	return true; //verify_optimal_price();
}

bool CS_solver::solve_feasible()
{
	long long max_c = initialize();

	if (max_c < 0)
		return false;

	cs2_solve_feasible(
		long(num_nodes()), long(num_arcs()),
		&nodes_[0], &arcs_[0],
		12, max_c);

	collect_everything();

	return true; //verify_optimal_price();
}

void CS_solver::collect_everything()
{
	cost_ = 0;

	for (size_t i = 0; i < num_arcs(); ++i)
	{
		if (backward_arc(i))
			continue;

		size_t i_edge = arcs_[i].arc_id-1;

		long flow = get_edge_ub(i_edge)-arcs_[i].r_cap;

		set_edge_flow(i_edge, flow);

		cost_ += get_edge_cost(i_edge)*flow;
	}
}

void CS_solver::collect_edge_flows(long f[]) const
{
	for (size_t i = 0; i < num_edges(); ++i)
		f[i] = get_edge_flow(i);
}

/*
void CS_solver::collect_vertex_potential(long p[])
{
	if (!potential_computed_)
	{
		compute_prices();
		potential_computed_ = true;
	}

	long long min_p = nodes_[0].price;

	for (size_t i = 1; i < num_nodes(); ++i)
		min_p = std::min(min_p, nodes_[i].price);

	for (size_t i = 0; i < num_nodes(); ++i)
		p[i] = long(nodes_[i].price-min_p);
}
//*/

bool CS_solver::verify_optimal_price()
{
	compute_prices();

	for (size_t i = 0; i < num_arcs(); ++i)
	{
		if (backward_arc(i))
			continue;

		arc_t *this_arc = &arcs_[i];
		arc_t *sis_arc = this_arc->sister;

		node_t *u = sis_arc->head;
		node_t *v = this_arc->head;

		size_t i_edge = this_arc->arc_id-1;

		long long rc = u->price+this_arc->cost-v->price;

		if (((this_arc->r_cap == 0) || (rc >= 0))
			&& ((sis_arc->r_cap == 0) || (rc <= 0)))
			continue;

		fprintf(stderr,
			"CS2: warning: %u (%u->%u) non-optimal, "
			"fw-rc %d, fw %d, bw %d\n",
			i_edge, u-&nodes_[0], v-&nodes_[0],
			(long)rc, this_arc->r_cap, sis_arc->r_cap);

		return false;
	}

	return true;
}

} // namespace flow
