#include "../misc/common.h"
#include "../misc/trace.h"
#include "../misc/timer.h"
#include "../misc/options.h"

#include "lr_sizing.h"

using namespace std;

#define TOL 1e-6

RegisterOption(lr_step, "1", "step size for subgradient optimization")
RegisterOption(lr_sechen, "false", "use sechen's method instead of standard one for LR")

lr_sizing::lr_sizing()
	: step_(misc::options::get().as_double("lr_step")),
	sechen_(misc::options::get().as_bool("lr_sechen"))
{
	fprintf(stdout, "@lr_step %e\n", step_);
}

/*
void lr_sizing::set_multipliers(double step)
{
	TraceScope("LR", 8);

//	double subgradient_norm = vars_->compute_subgradient_norm();
	
	for (size_t i = 0; i < vars_->num_edges(); ++i)
	{
		double previous_mp = vars_->get_mp(i);
		double slack = 0.0;

		if (step == 0)
		{
			if(iter > 0) slack = vars_->set_mp_slack_nostep(i);
			else slack = vars_->set_mp_slack(i, step+init , subgradient_norm);
		}
		else
		{
			slack = vars_->set_mp_slack(i, step/iter, subgradient_norm);
			else slack = vars_->set_mp_slack(i,step, subgradient_norm);
			double current_mp = vars_->get_mp(i);
			if(TraceActive()){
				fprintf(fpt,"Edge : %d, prev_mp : %e, slack : %e, current_mp : %e\n", i, previous_mp, slack, current_mp);
			}
		}
	}
	if(TraceActive())
		fclose(fpt);
	return;
}
*/

bool lr_sizing::find_next_flow(int round)
{
	//Algorithm to solve dual problem
	//1. Select any arbitrary set of multipliers
	//2. Solve Lagrangian subproblem
	//3. Update the lagrange multipliers based on subgradient optimization
	//4. Project \lambda such that optimality condition is satisfied
	//5. Repeat 2-4 till convergence
	
	// we assume before this loop 2 is done so start from 3, then 4, then 2

	if (sechen_)
		update_flow_by_sechen();
	else
		update_flow_by_slack(step_/(1+round));

	project_flow();

	update_flow(0);

	return true;
}

void lr_sizing::update_flow_by_slack(double step)
{
	const sizing_graph *sg = vars_->get_sizing_graph();
	size_t eOI = sg->get_clock_period_edge(0);
	size_t O = sg->get_edge_from(eOI);

	for (size_t i = 0;  i < vars_->num_edges(); ++i)
	{
		if (i == eOI) // don't update flow on OI edge
			continue;

		size_t u = sg->get_edge_from(i);
		size_t v = sg->get_edge_to(i);

		if (v != O)
		{
			flow_[i] -= step*vars_->compute_slack(i);
		}
		else
		{
			double real_slack = sg->get_clock_period()
				-vars_->get_delay(i)-vars_->get_at(u);
			flow_[i] -= step*real_slack;
		}
		if (flow_[i] < 0)
			flow_[i] = 0;
	}
}

void lr_sizing::update_flow_by_sechen()
{
	const sizing_graph *sg = vars_->get_sizing_graph();
	size_t eOI = sg->get_clock_period_edge(0);
	size_t O = sg->get_edge_from(eOI);

	for (size_t i = 0;  i < vars_->num_edges(); ++i)
	{
		if (i == eOI) // don't update flow on OI edge
			continue;

		size_t u = sg->get_edge_from(i);
		size_t v = sg->get_edge_to(i);
		double au = vars_->get_at(u);
		double av = vars_->get_at(v);
		double d = vars_->get_delay(i);

		if (v != O)
		{
			if (av-d > TOL) 
				flow_[i] *= au/(av-d);
		}
		else
		{
			flow_[i] *= au/sg->get_clock_period();
		}

		if (flow_[i] < 0)
			flow_[i] = 0;
	}
}

void lr_sizing::project_flow()
{
	const sizing_graph *sg = vars_->get_sizing_graph();

	size_t eOI = sg->get_clock_period_edge(0);
	size_t O = sg->get_edge_from(eOI);
	size_t I = sg->get_edge_to(eOI);

	if ((sg->num_clock_period_edges() > 1)
		|| (O != sg->num_vertices()-1)
		|| (I != 0))
	{
		fprintf(stderr,
			"%s: error: proportional multiplier projection "
			"will fail for multiple clock period edges\n",
			sg->get_graph_name().c_str());

		exit(0);
	}

	TraceScope("GTSIZING", 5);

	TracePrint("Printing Node Imbalances Before Projection\n");

	for(size_t i = O; i > 0;)
	{
		--i;

		double mp_fanin_sum = 0.0, mp_fanout_sum = 0.0;
		size_t num_fanins = 0, num_fanouts = 0;

		for (sizing_graph::edge_range r = sg->fanins(i);
			r.first != r.second; ++r.first)
		{
			size_t edge = sg->this_edge(r.first);
			assert(sg->get_edge_to(edge) == i);
			mp_fanin_sum += flow_[edge];
			++num_fanins;
		}
		for (sizing_graph::edge_range r = sg->fanouts(i);
			r.first != r.second; ++r.first)
		{
			size_t edge = sg->this_edge(r.first);
			assert(sg->get_edge_from(edge) == i);
			mp_fanout_sum += flow_[edge];
			++num_fanouts;
		}

		TracePrint("Node : %d, mp_fanin_sum : %e, mp_fanout_sum : %e\n",
			i, mp_fanin_sum, mp_fanout_sum);

		// distribute fanouts to fanins
		if (mp_fanin_sum < TOL)
		{
			// evenly redistribute
			double mp_avg = mp_fanout_sum/num_fanins;
			for (sizing_graph::edge_range r = sg->fanins(i);
				r.first != r.second; ++r.first)
			{
				size_t edge = sg->this_edge(r.first);
				flow_[edge] = mp_avg;
			}
		}
		else
		{
			// proportional redistribute
			double mp_ratio = mp_fanout_sum/mp_fanin_sum;
			for (sizing_graph::edge_range r = sg->fanins(i);
				r.first != r.second; ++r.first)
			{
				size_t edge = sg->this_edge(r.first);
				flow_[edge] *= mp_ratio;
			}
		}
	}

	{
	TraceScope("GTSIZING", 4);

	TracePrint("Printing Projected Node Imbalances After Projection\n");
	
	if (TraceActive())
	{
	  for(size_t i = 0; i < vars_->num_vertices(); ++i)
	    {
	      double mp_fanin_sum = 0.0, mp_fanout_sum = 0.0;
	      
	      for (sizing_graph::edge_range r = sg->fanins(i);
		   r.first != r.second; ++r.first)
		{
		  size_t edge = sg->this_edge(r.first);
		  mp_fanin_sum += flow_[edge];
		}
	      for (sizing_graph::edge_range r = sg->fanouts(i);
		   r.first != r.second; ++r.first)
		{
		  size_t edge = sg->this_edge(r.first);
		  mp_fanout_sum += flow_[edge];
		}
	      
	      if (fabs(mp_fanin_sum-mp_fanout_sum) > TOL)
		fprintf(stdout,
			"Node : %lu, mp_fanin_sum : %e, mp_fanout_sum : %e, Imbalance : %e\n",
			i, mp_fanin_sum, mp_fanout_sum, mp_fanin_sum-mp_fanout_sum);
	    }
	}
	}
}
