#include "../misc/common.h"


#include "../misc/trace.h"


#include "../misc/timer.h"


#include "../misc/options.h"


#include "lr_sizing.h"


#include "flow_sizing.h"





using namespace std;





#define TOL 1e-6





RegisterTraceScope(LR, 1, "show dual maximizing details")


RegisterOption(lr_step,"1","step size param for subgradient optimization")





lr_sizing::lr_sizing()


	: vars_(0), before_LRS_(0), total_iters_(0), feasible_(true)


{


}





double lr_sizing::compute_cost() const


{


	return vars_->compute_cost();


}





void lr_sizing::attach_vars(sizing_vars *p_vars)


{


	vars_ = p_vars;


}





void lr_sizing::set_multipliers(double w)


{


	for (size_t i = 0; i < vars_->num_edges(); ++i){


		vars_->set_mp(i,w);


	}


}


void lr_sizing::set_multipliers(double step, int iter)


{


	TraceScope("LR",8);


	FILE * fpt = NULL;


	


	double init = 1.0;





	if(TraceActive()){


		char filename[100];


		sprintf(filename,"multipliers%d.dump",iter);


		fpt = fopen(filename,"w");


	}





	double subgradient_norm = vars_->compute_subgradient_norm();


	


	for (size_t i = 0; i < vars_->num_edges(); ++i){


		double previous_mp = vars_->get_mp(i);


		double slack = 0.0;


		if(step == 0){


			if(iter > 0) slack = vars_->set_mp_slack_nostep(i);


			else slack = vars_->set_mp_slack(i, step+init , subgradient_norm);


		}


		else{


			if(iter > 0) slack = vars_->set_mp_slack(i, step/iter, subgradient_norm);


			else slack = vars_->set_mp_slack(i,step, subgradient_norm);


			double current_mp = vars_->get_mp(i);


			if(TraceActive()){


				fprintf(fpt,"Edge : %d, prev_mp : %e, slack : %e, current_mp : %e\n", i, previous_mp, slack, current_mp);


			}


		}


	}


	if(TraceActive())


		fclose(fpt);


	return;


}





bool lr_sizing::solve_fast_dual(sizing_vars *p_vars)


{


	TraceScope("LR", 8);





	attach_vars(p_vars);





	feasible_ = true;





	//Gradient based preprocessing stage for solving the dual problem


	//efficiently


	//TO DO : Implement the approach for comparison





	return feasible_;


}





bool lr_sizing::solve_dual(sizing_vars *p_vars)


{


	attach_vars(p_vars);





	feasible_ = true;


	


	int max_iter = misc::options::get().as_int("dual_rounds");


	double step = misc::options::get().as_double("lr_step");


	int i = 0;





	os::timer_t t;





	//Algorithm to solve dual problem


	//1. Select any arbitrary set of multipliers


	//2. Solve Lagrangian subproblem


	//3. Update the lagrange multipliers based on subgradient optimization


	//4. Project \lambda such that optimality condition is satisfied


	//5. Repeat 2-4 till convergence


	


	//1. Initial set of multipliers: all 0 satisfies KKT optimality


	set_multipliers(0);


	//Do one iteration of flow


	


	//bool exact = options::get().as_bool("exact_line");


	flow_sizing fs;


	bool exact = false;


	fs.solve_dual(p_vars, 2, exact);


	i++;


	


	TraceScope("LR",8);


	if(TraceActive()){


		char filename[100];


		sprintf(filename,"sizing_vars_debug_init.dump");


		FILE * fpt = fopen(filename,"w");	


		vars_->dump(fpt);


		fclose(fpt);


	}


	size_t iters;


	fprintf(stderr,"step size : %e\n",step);





	//steps 2 - 4 till convergence 


	for (; i < max_iter; ++i)


	{


		before_LRS_ = vars_->get_cost();


		feasible_ = (vars_->get_cost() <= vars_->get_cost_ub());





		iters = vars_->solve_LRS();


		total_iters_ += iters;


		after_LRS_ = vars_->get_cost();	


		


		//update the arrival times


		vars_->update_arrival_times(i);


		set_multipliers(step,i);


		vars_->project_multipliers(i);


		


		TraceScope("LR",8);


		if(TraceActive()){


			char filename[100];


			sprintf(filename,"sizing_vars_debug_%d.dump",i);


			FILE * fpt = fopen(filename,"w");


			vars_->dump(fpt);


			fclose(fpt);


		}


		t.stop();





		fprintf(stdout,


			"dual round %d, cost %e, target %e, Achieved delay %e, iters %u, "


			"total_iters %u, duality gap %e, real %.3f\n",


			i, vars_->get_cost(), vars_->get_target(), vars_->get_achieved_delay(),


			iters, total_iters_, vars_->get_fd(), t.real());


			{


				TraceScope("LR",1);


				TracePrint(


					"    before LRS %e, check cost %e\n",


					before_LRS_,


					fabs(vars_->compute_cost()-vars_->get_cost()));


			}


		if (!feasible_)


			break;


	}


	


	if (feasible_)


	{


		fprintf(stdout,


			"final feasible, rounds %d, cost %e, target %e, "


			"iters %u, total_iters %u\n",


			i, vars_->get_cost(), vars_->get_target(),


			iters, total_iters_);


	}


	else


	{


		fprintf(stdout,


			"final infeasible, rounds %d, cost %e, target %e, "


			"cost_ub %e, iters %u, total_iters %u\n",


			i, vars_->get_cost(), vars_->get_target(),


			vars_->get_cost_ub(), iters, total_iters_);


	}





	TracePrint("    before LRS %e, check cost %e\n",


		before_LRS_, fabs(vars_->compute_cost()-vars_->get_cost()));


	


	FILE * fpt = fopen("var_lr_sizing.dump","w");


	vars_->dump(fpt);


	fclose(fpt);


	


	return feasible_;


}


