#include "../misc/common.h"

#include "posynomial.h"

namespace math
{

size_t  posynomial::num_terms() const
{
	return terms_.size();
}

posynomial::sum_terms &posynomial::get_terms()
{
	return terms_;
}

const posynomial::sum_terms &posynomial::get_terms() const
{
	return terms_;
}

posynomial::sum_term_range posynomial::get_term_range()
{
	return sum_term_range(terms_.begin(), terms_.end());
}

posynomial::sum_term_crange posynomial::get_term_range() const
{
	return sum_term_crange(terms_.begin(), terms_.end());
}

double posynomial::evaluate(const double values[]) const
{
	double ret = 0;

	for (sum_term_crange cr = get_term_range();
		cr.first != cr.second; ++cr.first)
		ret += cr.first->evaluate(values);

	return ret;
}

std::string posynomial::to_string(
	misc::cstr_t var_prefix,
	misc::cstr_t format) const
{
	std::string ret;

	if (num_terms() == 0)
		ret = "0";
	else
		for (sum_term_crange cr = get_term_range();
			cr.first != cr.second; ++cr.first)
			ret += cr.first->to_string(var_prefix, format);

	return ret;
}

std::string posynomial::to_string(
	const misc::cstr_t var_names[],
	misc::cstr_t format) const
{
	std::string ret;

	if (num_terms() == 0)
		ret = "0";
	else
		for (sum_term_crange cr = get_term_range();
			cr.first != cr.second; ++cr.first)
			ret += cr.first->to_string(var_names, format);

	return ret;
}

double posynomial::sum_term::evaluate(const double values[]) const
{
	double ret = sum_coef;

	for (size_t i = 0; i < prod_terms.size(); ++i)
	{
		const product_term &pt = prod_terms[i];
		ret *= pow(values[pt.var_name], pt.pow_coef);
	}

	return ret;
}

std::pair<double, int> posynomial::sum_term::partial_evaluate(
	const double values[],
	size_t which_var) const
{
	int pow_coef = 0;
	double coef = sum_coef;

	for (size_t i = 0; i < prod_terms.size(); ++i)
	{
		const product_term &pt = prod_terms[i];

		if (pt.var_name == which_var)
			pow_coef += pt.pow_coef;
		else
			coef *= pow(values[pt.var_name], pt.pow_coef);
	}

	return std::make_pair(coef, pow_coef);
}

std::string posynomial::sum_term::to_string(
	misc::cstr_t var_prefix,
	misc::cstr_t format) const
{
	char format_str[100];

	char temp[1000];
	sprintf(format_str, "%%+%s", format);
	sprintf(temp, format_str, sum_coef);

	std::string ret(temp);

	for (size_t i = 0; i < prod_terms.size(); ++i)
	{
		const product_term &pt = prod_terms[i];

		sprintf(temp, " %s_{%lu}^{%d}", var_prefix, pt.var_name, pt.pow_coef);
		ret += temp;
	}

	return ret;
}

std::string posynomial::sum_term::to_string(
	const misc::cstr_t var_names[],
	misc::cstr_t format) const
{
	char format_str[100];

	char temp[1000];
	sprintf(format_str, "%%+%s", format);
	sprintf(temp, format_str, sum_coef);

	std::string ret(temp);

	for (size_t i = 0; i < prod_terms.size(); ++i)
	{
		const product_term &pt = prod_terms[i];

		sprintf(temp, " %s^{%d}", var_names[pt.var_name], pt.pow_coef);
		ret += temp;
	}

	return ret;
}

void posynomial::sum_term::add_term(size_t var_name, int pow_coef)
{
	prod_terms.push_back(product_term(var_name, pow_coef));

}
posynomial::sum_term &posynomial::add_term(double sum_coef)
{
	terms_.push_back(sum_term(sum_coef));
	return terms_.back();
}

posynomial::sum_term::sum_term(double sc)
	: sum_coef(sc)
{
}

posynomial::product_term::product_term(size_t vn, int pc)
	: var_name(vn), pow_coef(pc)
{
}

void posynomial::save(FILE *fp) const
{
	fprintf(fp, "%lu ", num_terms());

	for (sum_term_crange cr = get_term_range();
		cr.first != cr.second; ++cr.first)
		cr.first->save(fp);

	fprintf(fp, "\n");
}

bool posynomial::load(FILE *fp)
{
	size_t n = 0;
	if (fscanf(fp, "%lu", &n) != 1)
		return false;

	for (size_t i = 0; i < n; ++i)
	{
		sum_term &st = add_term();

		if (!st.load(fp))
			return false;
	}

	return true;
}

void posynomial::sum_term::save(FILE *fp) const
{
	fprintf(fp, "%e %lu ", sum_coef, prod_terms.size());

	for (size_t i = 0; i < prod_terms.size(); ++i)
	{
		const product_term &pt = prod_terms[i];

		fprintf(fp, "%lu %d ", pt.var_name, pt.pow_coef);
	}
}

bool posynomial::sum_term::load(FILE *fp)
{
	size_t n = 0;
	if (fscanf(fp, "%lf %lu", &sum_coef, &n) != 2)
		return false;
	
	prod_terms.reserve(n);

	for (size_t i = 0; i < n; ++i)
	{
		int var_name, pow_coef;
		
		if (fscanf(fp, "%u %d", &var_name, &pow_coef) != 2)
			return false;

		add_term(var_name, pow_coef);
	}

	return true;
}

}; // namespace math
