#include "../misc/common.h"
#include "../misc/trace.h"

#include "sizing_vars.h"

#define TOL 1e-6

sizing_vars::sizing_vars()
	: sg_(0),
	target_(0),
	fd_(0),
	cost_ub_(0),
	use_T_2_(false),
	T_w_(0),
	T_(0),
	T_lb_(0),
	T_ub_(0)
//	cr_(0)
{
}

sizing_vars::~sizing_vars()
{
//	delete cr_;
}

void sizing_vars::attach_graph(const sizing_graph *p_sg)
{
  assert(!use_T_2_);
  
  sg_ = p_sg;
  
  // start with 0 as mps
  mps_.assign(num_edges(), 0);
  
  vars_.resize(num_vars());
  delays_.resize(num_edges());
  ats_.resize(num_vertices());
  
  Q_.assign(num_vars());
  
  A_.assign(num_vertices());
  B_.assign(num_vertices());
  topo_.assign(num_vertices());
  current_topo_edges_.resize(num_vertices());
  
  check_topo_order();
  
  initialize_vars();
  update_delays();
  
  cp_delays_.resize(num_edges());
  cp_weights_.assign(num_edges(), 0);
  
  cp_weights_[sg_->get_clock_period_edge(0)] = -1;
  
  if (sg_->is_skew_opt())
    {
      //		assert(cr_ == 0);
      
      for (size_t i = 0; i < sg_->num_dffs(); ++i)
	{
	  cp_weights_[sg_->get_DQ_edge(i)] = -1;
	  cp_weights_[sg_->get_IQ_edge(i)] = sg_->get_skew_lb();
	  cp_weights_[sg_->get_DO_edge(i)] = -sg_->get_skew_ub();
	}
      
      //		const graph2::graph_v *graph = sg_->get_graph();
      //		cr_ = new cycle_ratio(*graph,
      //			graph->make_edge_property(delays_wo_cp_),
      //			graph->make_edge_property(cp_weights_));
    }

  update_arrival_times();
}

void sizing_vars::check_topo_order() const
{
	for (size_t i = 0; i < num_edges(); ++i)
	{
		if (sg_->is_clock_period_edge(i))
			continue;

		size_t from = sg_->get_edge_from(i);
		size_t to = sg_->get_edge_to(i);

		if (to < from)
		{
			fprintf(stderr,
				"%s: error: edge %lu (%lu->%lu) violates "
				"pre-assumed partial order\n",
				sg_->get_graph_name().c_str(), i, from, to);

			exit(0);
		}
	}
}

int sizing_vars::update_var(size_t which_var, double value)
{
	assert(sg_ != 0);

	int ret = UPDATED;
	double real = value;
	double lb = sg_->get_var_lb(which_var);
	double ub = sg_->get_var_ub(which_var);
	if (value < lb)
		real = lb, ret = SET_MIN;
	else if (value > ub)
		real = ub, ret = SET_MAX;

	if (fabs(real-vars_[which_var]) < TOL)
		return NO_CHANGE;

	double w = sg_->get_var_weight(which_var);
	target_ -= vars_[which_var]*w;
	vars_[which_var] = real;
	target_ += vars_[which_var]*w;

	return ret;
}

void sizing_vars::set_mp(size_t which_edge, double mp)
{
	assert(which_edge < num_edges());

	double d = delays_[which_edge];
	fd_ -= mps_[which_edge]*d;
	mps_[which_edge] = mp;
	fd_ += mps_[which_edge]*d;
}

double sizing_vars::get_at(size_t which_vertex) const
{
	assert(which_vertex < num_vertices());

	return ats_[which_vertex];
}

double sizing_vars::compute_slack(size_t which_edge) const
{
	assert(which_edge < num_edges());

	size_t u = sg_->get_edge_from(which_edge);
	size_t v = sg_->get_edge_to(which_edge);
	double d = get_delay(which_edge);

	return get_at(v)-d-get_at(u);
}

const double *sizing_vars::get_vars() const
{
	assert(sg_ != 0);

	return &vars_[0];
}

const double *sizing_vars::get_delays() const
{
	assert(sg_ != 0);

	return &delays_[0];
}

const double *sizing_vars::get_mps() const
{
	assert(sg_ != 0);

	return &mps_[0];
}

void sizing_vars::initialize_vars()
{
	assert(sg_ != 0);

	target_ = 0;
	cost_ub_ = 0;

	for (size_t i = 0; i < num_vars(); ++i)
	{
		vars_[i] = sg_->get_var_lb(i);
		double w = sg_->get_var_weight(i);

		target_ += w*vars_[i];
		cost_ub_ += w*sg_->get_var_ub(i);
	}
}
	
void sizing_vars::update_arrival_times()
{
	const size_t N = num_vertices();

	for(size_t i = 0; i < N; i++)
	{
		double max_t = 0;

		for(sizing_graph::edge_range r = sg_->fanins(i);
			r.first != r.second; ++r.first)
		{
			size_t e = sg_->this_edge(r.first);

			if (sg_->is_clock_period_edge(e))
				continue;

			size_t u = sg_->get_edge_from(e);
			double t = delays_[e]+ats_[u];

			if (t > max_t) 
				max_t = t;	
		}

		ats_[i] = max_t;

		assert(ats_[i] >= 0 && "arrival times can't be negative");
	}
}

void sizing_vars::update_delays()
{
	assert(sg_ != 0);

	fd_ = 0;
	for (size_t i = 0; i < num_edges(); ++i)
	{
		double d = sg_->get_edge_delay(i).evaluate(get_vars());

		if (use_T_2_ && sg_->is_clock_period_edge(i))
		{
			d = -T_;
		}

		delays_[i] = d;
		fd_ += mps_[i]*d;
	}
}

double sizing_vars::get_cost() const
{
	return get_target()+fd_;
}

double sizing_vars::get_target() const
{
	if (use_T_2_)
		return target_+T_w_*T_*T_;
	else
		return target_;
}

double sizing_vars::get_fd() const
{
	return fd_;
}

double sizing_vars::get_cost_ub() const
{
	if (is_T_2_used())
		return cost_ub_+T_w_*T_ub_*T_ub_;
	else
		return cost_ub_;
}

double sizing_vars::compute_cost() const
{
	assert(!use_T_2_);
	assert(sg_ != 0);

	double target = std::inner_product(
		vars_.begin(), vars_.end(), sg_->get_var_weights(),	0.0);

	double fd = compute_fd();

	return target+fd;
}

double sizing_vars::compute_fd() const
{
	assert(sg_ != 0);

	double fd = std::inner_product(
		delays_.begin(), delays_.end(), mps_.begin(), 0.0);

	return fd;
}

size_t sizing_vars::solve_LRS()
{
	assert(sg_ != 0);

	TraceScope("GTSIZING", 8);

	pre_solve_LRS();

	Q_.insert_all();

	size_t iter = 0;

	for (; !Q_.empty(); ++iter)
	{
		size_t v = Q_.pop_front();

		if (!LRS_update_var(v))
			continue;

		typedef sizing_graph::id_crange id_crange;
		for (id_crange cr = sg_->get_related_edges(v);
			cr.first != cr.second; ++cr.first)
			for (id_crange cer = sg_->get_related_vars(*cr.first);
				cer.first != cer.second; ++cer.first)
			{
				size_t u = *cer.first;

				if ((u != v) && !Q_.in_list(u))
					Q_.push_back(u);
			}
	}

	if (TraceActive())
		dump(stdout);

	post_solve_LRS();

	return iter;
}

void sizing_vars::pre_solve_LRS()
{
}

void sizing_vars::post_solve_LRS()
{
	if (use_T_2_)
		update_T();
	update_delays();
	update_arrival_times();
}

void sizing_vars::compute_cp_delays(double cp)
{
	double dT = cp-sg_->get_clock_period();

	for (size_t i = 0; i < num_edges(); ++i)
		cp_delays_[i] = delays_[i]+dT*cp_weights_[i];
}

bool sizing_vars::feasible_cp_delays(double cp)
{
	compute_cp_delays(cp);

	// Goldberg - Radzik

	// assume a pseudo node s=n connecting to every vertex with weight 0
	std::fill(ats_.begin(), ats_.end(), 0);
	A_.erase_all();
	B_.insert_all();
	topo_.erase_all();

	for (size_t r = 0; !B_.empty(); ++r)
	{
		fprintf(stdout, "GR: round %lu, B %lu\n", r, B_.size());

		// compute A from B
		assert(A_.empty());
		for (; !B_.empty();)
		{
			size_t u = B_.pop_front();

			if (A_.in_list(u))
				continue; // nothing new, already in A after previous topo

			sizing_graph::edge_range er = sg_->fanouts(u);
			for (; er.first != er.second; ++er.first)
			{
				size_t e = sg_->this_edge(er.first);
				size_t v = sg_->get_edge_to(e);
				assert(u == sg_->get_edge_from(e));

				double atv = ats_[u]+cp_delays_[e];

				if (atv > ats_[v])
					break;
			}
			if (er.first == er.second)
				continue; // no fanout positive edge

			// changes pending

			assert(topo_.empty());
			topo_.push_back(u);
			current_topo_edges_[u] = sg_->fanouts(u).first;

			for (; !topo_.empty();)
			{
				size_t u = topo_.back();

				for (sizing_graph::edge_range er = sg_->fanouts(u);
					current_topo_edges_[u] != er.second;
					++current_topo_edges_[u])
				{
					size_t e = sg_->this_edge(current_topo_edges_[u]);
					size_t v = sg_->get_edge_to(e);
					assert(u == sg_->get_edge_from(e));

					double atv = ats_[u]+cp_delays_[e];

					// a negative edge
					if (atv < ats_[v]-0.000001)
						continue;

					// already there
					if (A_.in_list(v))
						continue;

					// positive cycle found
					if (topo_.in_list(v))
						return false;

					// continue DFS
					++current_topo_edges_[u];
					topo_.push_back(v);
					current_topo_edges_[v] = sg_->fanouts(v).first;
					break;
				}

				if (topo_.back() == u) // nothing new?
				{
					topo_.pop_back();
					A_.push_front(u);
				}
			}
		}

		// BF
		for (; !A_.empty();)
		{
			size_t u = A_.pop_front();

			for (sizing_graph::edge_range er = sg_->fanouts(u);
				er.first != er.second; ++er.first)
			{
				size_t e = sg_->this_edge(er.first);
				size_t v = sg_->get_edge_to(e);
				assert(u == sg_->get_edge_from(e));

				double atv = ats_[u]+cp_delays_[e];

				if (atv > ats_[v])
				{
					ats_[v] = atv;

					if (!B_.in_list(v) && !A_.in_list(v))
					{
						B_.push_back(v);
					}
				}
			}
		}
	}

	return true;
}

double sizing_vars::get_achieved_clock_period(bool skew_exact)
{
	assert(sg_ != 0);

	if (!sg_->is_skew_opt())
	{
		assert(sg_->num_clock_period_edges() == 1);

		size_t eoi = sg_->get_clock_period_edge(0);
		
		size_t po = sg_->get_edge_from(eoi);
		size_t pi = sg_->get_edge_to(eoi);

		return ats_[po]-ats_[pi];
	}
	else if (!skew_exact) // return -1 if feasible, otherwise return 2T
	{
		double T = sg_->get_clock_period();

		return feasible_cp_delays(T)? -1: 2*T;
	}
	else
	{
		double T = sg_->get_clock_period();

		double lb = 0, ub = T;

		for (;; ub = ub*2)
		{
			if (feasible_cp_delays(ub))
				break;
		}

		fprintf(stdout, "skew-opt: info: clock period in [%f,%f]\n", lb, ub);

		for (; ub-lb > 1e-4*T;)
		{
			double mid = (lb+ub)/2;

			if (feasible_cp_delays(mid))
				ub = mid;
			else
				lb = mid;

			fprintf(stdout, "skew-opt: info: clock period in [%f,%f]\n", lb, ub);
		}

		return ub;
	}
}

std::pair<double, double> sizing_vars::compute_coef_1m1(size_t which_var)
{
	double coef_1 = sg_->get_var_weight(which_var);
	double coef_m1 = 0;

	for (sizing_graph::id_crange cr = sg_->get_related_edges(which_var);
		cr.first != cr.second; ++cr.first)
	{
		size_t e = *cr.first;
		const math::posynomial &dp = sg_->get_edge_delay(e);

		for (math::posynomial::sum_term_crange ctr = dp.get_term_range();
			ctr.first != ctr.second; ++ctr.first)
		{
			std::pair<double, int> result
				= ctr.first->partial_evaluate(get_vars(), which_var);

			switch (result.second)
			{
			case 0:
				break;
			case 1:
				coef_1 += mps_[e]*result.first;
				break;
			case -1:
				coef_m1 += mps_[e]*result.first;
				break;
			default:
				fprintf(stderr, "error: wrong order of var %lu\n", which_var);
				exit(0);
			}
		}
	}

	return std::make_pair(coef_1, coef_m1);
}

bool sizing_vars::LRS_update_var(size_t which_var)
{
	TraceScope("GTSIZING", 9);
	
	std::pair<double, double> coefs = compute_coef_1m1(which_var);

	double coef_1 = coefs.first;
	double coef_m1 = coefs.second;
	double value = sqrt(coef_m1/coef_1);

	if (fabs(coef_1) < 1e-6)
		value = sg_->get_var_ub(which_var);
	else if (coef_m1 < 0)
		value = sg_->get_var_lb(which_var);

	int action = update_var(which_var, value);

	misc::cstr_t actions[] = {"no change", "updated", "set min", "set max"};

	TracePrint(		
		"update var %lu: %s, sqrt(%e/%e)->%e, %s\n",
		which_var,
		sg_->get_var_name(which_var),
		coef_m1,
		coef_1,
		value,
		actions[action]);

	return action != NO_CHANGE;
}

size_t sizing_vars::solve_LRS2()
{
	assert(sg_ != 0);

	TraceScope("GTSIZING", 8);

	pre_solve_LRS();

	size_t iter = 0;
	
	for (;; ++iter)
	{
		bool update_one = false;
		for (size_t i = 0; i < num_vars(); ++i)
			if (LRS_update_var(i))
				update_one = true;

		if (!update_one)
			break;
	}

	if (TraceActive())
		dump(stdout);

	post_solve_LRS();

	return (iter+1)*num_vars();
}

size_t sizing_vars::num_edges() const
{
	assert(sg_ != 0);

	return sg_->num_edges();
}

size_t sizing_vars::num_vars() const
{
	assert(sg_ != 0);

	return sg_->num_vars();
}

size_t sizing_vars::num_vertices() const
{
	assert(sg_ != 0);

	return sg_->num_vertices();
}

double sizing_vars::get_delay(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return delays_[which_edge];
}

double sizing_vars::get_mp(size_t which_edge) const
{
	assert(which_edge < num_edges());

	return mps_[which_edge];
}

double sizing_vars::get_var(size_t which_var) const
{
	assert(which_var < num_vars());

	return vars_[which_var];
}

const sizing_graph *sizing_vars::get_sizing_graph() const
{
	return sg_;
}

void sizing_vars::dump(FILE *fp) const
{
	fprintf(fp, "+++++++++++++++++++++++++++++++++++++++++++++\n");
	for (size_t i = 0; i < num_vars(); ++i)
	{
		fprintf(fp,
			"var %lu: %s, %e, [%e, %e]\n",
			i,
			sg_->get_var_name(i),
			vars_[i],
			sg_->get_var_lb(i),
			sg_->get_var_ub(i));
	}
	for (size_t i = 0; i < num_edges(); ++i)
	{
		if (mps_[i] < TOL)
			continue;

		fprintf(fp,
			"edge %lu, mp %e, delay %e\n",
			i,
			mps_[i],
			delays_[i]);
	}
	fprintf(fp, "---------------------------------------------\n");
}

bool sizing_vars::is_T_2_used() const
{
	return use_T_2_;
}

double sizing_vars::get_T() const
{
	assert(use_T_2_);

	return T_;
}

void sizing_vars::set_T_2(double w, double lb, double ub)
{
	assert(w > 0);

	use_T_2_ = true;
	T_w_ = w;
	T_lb_ = lb;
	T_ub_ = ub;

	update_T();

	for (size_t i = 0; i < sg_->num_clock_period_edges(); ++i)
	{
		size_t e = sg_->get_clock_period_edge(i);
		delays_[e] = -T_;
	}

	fd_ = compute_fd();
}

void sizing_vars::update_T()
{
	assert(use_T_2_);

	TraceScope("GTSIZING", 5);

	double total_mps = 0;
	for (size_t i = 0; i < sg_->num_clock_period_edges(); ++i)
	{
		size_t e = sg_->get_clock_period_edge(i);
		total_mps += mps_[e];
	}

	double T = total_mps/2/T_w_;

	if (T < T_lb_)
		T_ = T_lb_;
	else if (T > T_ub_)
		T_ = T_ub_;
	else
		T_ = T;

	TracePrint(
		"update T: total mps %e, optimal %e, real %s\n",
		total_mps, T, T_);
}

double sizing_vars::get_T_2_dual_cost() const
{
	assert(use_T_2_);

	return target_+T_w_*T_*T_+fd_;
}

double sizing_vars::get_T_2_primal_cost()
{
	assert(use_T_2_);

	double T = get_achieved_clock_period();

	if (T < T_lb_) // should clip T
		T = T_lb_;

	return target_+T_w_*T*T;
}

double sizing_vars::get_T_2_original_cost() const
{
	return target_;
}
