#ifndef SIZING_VARS_H
#define SIZING_VARS_H

#include "../graph/graph_utilities.h"
//#include "../graph/howard.h"

#include "sizing_graph.h"

class sizing_vars
{
	sizing_vars(sizing_vars &);
	sizing_vars &operator = (const sizing_vars&);

public:
	sizing_vars();
	virtual ~sizing_vars();

	virtual void attach_graph(const sizing_graph *p_sg);
	const sizing_graph *get_sizing_graph() const;

	const double *get_vars() const;
	const double *get_delays() const;

	const double *get_mps() const;

	size_t num_edges() const;
	size_t num_vars() const;
	size_t num_vertices() const;

	double get_var(size_t which_edge) const;
	double get_delay(size_t which_edge) const;
	
	double get_mp(size_t which_edge) const;
	virtual void set_mp(size_t which_edge, double mp);

	void update_arrival_times();
	double get_achieved_clock_period(bool skew_exact = false);
	double get_at(size_t which_vertex) const;
	double compute_slack(size_t which_edge) const;

	void dump(FILE *fp) const;

	size_t solve_LRS();
	size_t solve_LRS2();

	double get_target() const;
	double get_fd() const;
	double get_cost() const;
	double get_cost_ub() const;
	double compute_fd() const;
	
	virtual double compute_cost() const;
	
	// use w T^2 in cost function
	bool is_T_2_used() const;
	void set_T_2(double w, double lb, double ub);
	double get_T() const;
	double get_T_2_dual_cost() const;
	double get_T_2_primal_cost();
	double get_T_2_original_cost() const;

protected:
	// will decide where to go and update the cache if needed
	enum {NO_CHANGE = 0, UPDATED, SET_MIN, SET_MAX};
	
	//Functions related to variable updates
	virtual int update_var(size_t which_var, double value);

	virtual void initialize_vars();
	
	virtual bool LRS_update_var(size_t which_var);
	
	//Extended for elmore delay formulation
	virtual std::pair<double, double> compute_coef_1m1(size_t which_var);
	
	//Delay and topology related functions
	void check_topo_order() const;

	void update_delays();
	
	virtual void pre_solve_LRS();
	virtual void post_solve_LRS();
	
	const sizing_graph *sg_;
	
	std::vector<double> vars_, delays_, mps_;

	//arrival time updates
	std::vector<double> ats_;

	double target_, fd_, cost_ub_;

	graph::fast_list Q_;

	// Goldberg - Radzik
	graph::fast_list A_, B_, topo_;
	std::vector<sizing_graph::out_edge_iterator> current_topo_edges_;

	// additional w T^2 term
	bool use_T_2_;
	double T_w_, T_, T_lb_, T_ub_;

	void update_T();

	// for clock period computation with skew
	std::vector<double> cp_delays_, cp_weights_;

	void compute_cp_delays(double cp);
	bool feasible_cp_delays(double cp);

//	typedef graph2::howard<
//		graph2::graph_v::edge_properties<double>::array_prop,
//		graph2::graph_v::edge_properties<double>::array_prop> cycle_ratio;
//	cycle_ratio *cr_;

}; // class sizing_vars

// speed up coef_1/coef_m1 computation in sizing_vars::LRS_update_var
// for Elmore delays by caching
class sizing_vars_elmore : public sizing_vars
{
public:
  virtual void attach_graph(const sizing_graph *p_sg);

  const double *get_inv_vars() const;

protected:
	virtual int update_var(size_t which_var, double value);
	virtual void pre_solve_LRS();
	virtual std::pair<double, double> compute_coef_1m1(size_t which_var);

	void build_coef_caches();
	void pre_compute_coef();

	std::vector<double> inv_vars_;

	class coef_cache
	{
	public:
		typedef std::pair<double, size_t> coef_index_pair;

		void add_const(double coef, size_t which_edge);
		void add_coef(double coef, size_t which_var, size_t which_edge);

		void update_cache(const double mps[]);

		double evaluate(const double vars[]);

	protected:
		double cached_const_;

		std::vector<coef_index_pair> const_mps_;
		std::vector<coef_index_pair> coef_vars_;
		std::vector<coef_index_pair> coef_mps_;

	}; // class coef_cache

	std::vector<coef_cache> coef_1s_, coef_m1s_;

}; // class sizing_vars_elmore

#endif // SIZING_VARS_H
