#include "../misc/common.h"


#include "../misc/trace.h"





#include "sizing_vars.h"





#define TOL 1e-6





RegisterTraceScope(LRS, 1, "show LRS details")





sizing_vars::sizing_vars()


	: sg_(0)


{


}





void sizing_vars::attach_graph(const sizing_graph *p_sg)


{


	sg_ = p_sg;





	vars_.resize(num_vars());


	delays_.resize(num_edges());


	mps_.assign(num_edges(), 0);


	ats_.resize(num_vertices());


	dats_.resize(num_vertices());





	initialize_vars();


	update_delays();


	initialize_ats();


	identify_po_pi();


}





int sizing_vars::update_var(size_t which_var, double value)


{


	assert(sg_ != 0);





	int ret = UPDATED;


	double real = value;


	double lb = sg_->get_var_lb(which_var);


	double ub = sg_->get_var_ub(which_var);


	if (value < lb)


		real = lb, ret = SET_MIN;


	else if (value > ub)


		real = ub, ret = SET_MAX;





	if (fabs(real-vars_[which_var]) < TOL)


		return NO_CHANGE;





	double w = sg_->get_var_weight(which_var);


	target_ -= vars_[which_var]*w;


	vars_[which_var] = real;


	target_ += vars_[which_var]*w;





	return ret;


}





void sizing_vars::set_mp(size_t which_edge, double mp)


{


	assert(which_edge < num_edges());





	double d = delays_[which_edge];


	fd_ -= mps_[which_edge]*d;


	mps_[which_edge] = mp;


	fd_ += mps_[which_edge]*d;


}





//set multiplies based on slack but independent of step size


double sizing_vars::set_mp_slack_nostep(size_t which_edge)


{


	assert(which_edge < num_edges());


	typedef sizing_graph::id_crange id_crange;


	size_t u = sg_->get_edge_from(which_edge);


	size_t v = sg_->get_edge_to(which_edge);


	double d = get_delay(which_edge);


	double step = 0.0;


	//fprintf(stderr, "AT[u] : %e, AT[v] : %e, Delay : %e, mps : %e \n",ats_[u],ats_[v],d, mps_[which_edge]);





	if( ats_[v] - d > 0 )


		step = ats_[u]/(ats_[v]-d);





	double new_mp = step*mps_[which_edge];


	set_mp(which_edge,new_mp);


	return step;


}





//set multipliers based on slacks


//should be called after LRS has been done and multipliers are updated


double sizing_vars::set_mp_slack(size_t which_edge, double step, double norm)


{


	assert(which_edge < num_edges());





	typedef sizing_graph::id_crange id_crange;


	size_t u = sg_->get_edge_from(which_edge);


	size_t v = sg_->get_edge_to(which_edge);


	double d = get_delay(which_edge);





	double slack = (ats_[u] + d) - ats_[v];


	


	double new_mp = mps_[which_edge] + step*(-slack);





	if(new_mp < 0) new_mp = 0;


	


	set_mp(which_edge,new_mp);


	


	return slack;


}





double sizing_vars::compute_subgradient_norm() const


{


	double norm = 0.0;


	for(size_t i = 0; i < num_edges(); i++){


		size_t u = sg_->get_edge_from(i);


		size_t v = sg_->get_edge_to(i);


		double slack = ats_[u] + get_delay(i) - ats_[v];


		norm += slack*slack;


	}


	return sqrt(norm);


}





void sizing_vars::project_multipliers(int iter)


{


	TraceScope("LR",8);


	if(TraceActive())


		fprintf(stderr,"Printing Node Imbalances\n");





	typedef sizing_graph::id_crange id_crange;


	for(size_t i = 0; i < num_vertices(); i++){


		id_crange cr = sg_->get_fanin_edges(i);


		double mp_fanin_sum = 0.0, mp_fanout_sum = 0.0;


		


		size_t num_fanins = 0;


		for(;cr.first != cr.second; ++cr.first){


			size_t edge = *cr.first;


			mp_fanin_sum += get_mp(edge);


			num_fanins += 1;


		}


		size_t num_fanouts = 0;


		cr = sg_->get_fanout_edges(i);


		for(;cr.first != cr.second; ++cr.first){


			size_t edge = *cr.first;


			mp_fanout_sum += get_mp(edge);


			num_fanouts += 1;


		}


		TraceScope("LR",8);


		if(TraceActive())


			fprintf(stderr,"Node : %d, mp_fanin_sum : %e, mp_fanout_sum : %e\n",i,mp_fanin_sum,mp_fanout_sum);


		


		//case 1 : Sum of fanin multipliers equal to 0 and sum of fanout multipliers is


		//nonzero. Distribute fanin multipliers equally to fanout edges


		if( mp_fanin_sum == 0 && mp_fanout_sum != 0){


			cr = sg_->get_fanout_edges(i);


			for(;cr.first != cr.second; ++cr.first){


				size_t edge = *cr.first;


				set_mp(edge, 0);


			}


		}


		//case 2 : Sum of fanin multipliers not equal to 0, but fanout multipliers equal to 0


		//assign fanin multipliers as 0 


		else if( mp_fanin_sum != 0 && mp_fanout_sum == 0){


			double mp_avg = mp_fanin_sum/num_fanouts;


			cr = sg_->get_fanout_edges(i);


			for(; cr.first != cr.second; ++cr.first){


				size_t edge = *cr.first;


				set_mp(edge,mp_avg);


			}


		}


		//case 3 : Both sum of fanin and fanout multipliers are 0. No need to do any projection


		else if (mp_fanin_sum == 0 && mp_fanout_sum == 0){


			continue;


		}


		else{


			std::vector<double> percentage;


			percentage.resize(num_fanouts);


			num_fanouts = 0;


		


			cr = sg_->get_fanout_edges(i);


			for(;cr.first != cr.second; ++cr.first){


				size_t edge = *cr.first;


				percentage[num_fanouts] = get_mp(edge)/mp_fanout_sum;


				num_fanouts++;


			}


			num_fanouts = 0;


			cr = sg_->get_fanout_edges(i);


			for(;cr.first != cr.second; ++cr.first){


				size_t edge = *cr.first;


				double mp = percentage[num_fanouts]*mp_fanin_sum;


				set_mp(edge,mp);


				num_fanouts++;


			}


		}


	}


	if(TraceActive())


		fprintf(stderr,"Printing Projected Node Imbalances\n");


	


	for(size_t i = 0; i < num_vertices(); ++i)


	{


		id_crange cr = sg_->get_fanin_edges(i);


		double mp_fanin_sum = 0.0, mp_fanout_sum = 0.0;


		size_t num_fanins = 0;


		for(;cr.first != cr.second; ++cr.first){


			size_t edge = *cr.first;


			mp_fanin_sum += get_mp(edge);


			num_fanins++;


		}


		size_t num_fanouts = 0;


		cr = sg_->get_fanout_edges(i);


		for(;cr.first != cr.second; ++cr.first){


			size_t edge = *cr.first;


			mp_fanout_sum += get_mp(edge);


			num_fanouts++;


		}


		TraceScope("LR",8);


		if(TraceActive()){


			fprintf(stderr,"Node : %d, mp_fanin_sum : %e, mp_fanout_sum : %e\n",i,mp_fanin_sum,mp_fanout_sum);


		}


		else


		{


			assert(fabs(mp_fanin_sum-mp_fanout_sum) < TOL && "node imbalance is not zero");


		}


	}





	if(TraceActive()){


		char filename[100];


		sprintf(filename,"projected_multipliers%d.dump",iter);


		FILE * fpt = fopen(filename,"w");


		for (size_t i = 0; i < num_edges(); ++i){


			double projected_mp = get_mp(i);


			fprintf(fpt,"Edge : %d, current_mp : %e\n", i, projected_mp);


		}


		fclose(fpt);


	}


	


	return;


}


		


const double *sizing_vars::get_vars() const


{


	assert(sg_ != 0);





	return &vars_[0];


}





const double *sizing_vars::get_delays() const


{


	assert(sg_ != 0);





	return &delays_[0];


}





const double *sizing_vars::get_mps() const


{


	assert(sg_ != 0);





	return &mps_[0];


}





void sizing_vars::initialize_vars()


{


	assert(sg_ != 0);





	target_ = 0;


	cost_ub_ = 0;





	for (size_t i = 0; i < num_vars(); ++i)


	{


		vars_[i] = sg_->get_var_lb(i);





		double w = sg_->get_var_weight(i);





		target_ += w*vars_[i];


		cost_ub_ += w*sg_->get_var_ub(i);


	}


}





void sizing_vars::initialize_ats()


{


	assert(sg_ != 0);


	update_arrival_times(0);


}





void sizing_vars::identify_po_pi()


{


	for (size_t i = 0; i < num_edges(); ++i){


		if(delays_[i] < 0){


			po_ = sg_->get_edge_from(i);


			pi_ = sg_->get_edge_to(i);


		}


	}


}


	


//topological traversal to update the arrival times


void sizing_vars::update_arrival_times(int iter)


{


	const size_t N = num_vertices();


	for(size_t i = 0; i < N; i++){


		double max = 0.0;


		typedef sizing_graph::id_crange id_crange;


		id_crange cr = sg_->get_fanin_edges(i);


		for(; cr.first != cr.second; ++cr.first){


			size_t edge = *cr.first;


			double d = delays_[edge];


			if(d < 0) continue;


			size_t u = sg_->get_edge_from(edge);


			double temp = d + ats_[u];


			if(temp > max) 


				max = temp;	


		}


		//computes delta change in arrival time


		dats_[i] = ats_[i] - max;


		//updates the new arrival time


		ats_[i] = max;


		assert(ats_[i] >= 0 && "arrival times can't be negative");


	}


	TraceScope("LR",8);


	if(TraceActive()){


		char filename[100];


		sprintf(filename,"arrival_times%d.dump",iter);


		FILE * fpt = fopen(filename,"w");


		for(size_t i = 0; i < N; i++)


			fprintf(fpt,"Port : %s, AT : %e, Delta AT : %e\n",sg_->get_vertex_name(i),ats_[i], dats_[i]);


		fclose(fpt);


		


		sprintf(filename,"delays%d.dump",iter);


		fpt = fopen(filename,"w");


		for(size_t i = 0; i < num_edges(); i++)


			fprintf(fpt,"Edge : %d, delay : %e\n",i,delays_[i]);


		fclose(fpt);


	}


}





void sizing_vars::update_delays()


{


	assert(sg_ != 0);





	fd_ = 0;


	for (size_t i = 0; i < num_edges(); ++i)


	{


		double d = sg_->get_edge_delay(i).evaluate(get_vars());


		delays_[i] = d;


		fd_ += mps_[i]*d;


	}


}





double sizing_vars::get_cost() const


{


	return target_+fd_;


}





double sizing_vars::get_target() const


{


	return target_;


}





double sizing_vars::get_fd() const


{


	return fd_;


}





double sizing_vars::get_cost_ub() const


{


	return cost_ub_;


}





double sizing_vars::compute_cost() const


{


	assert(sg_ != 0);





	double target = std::inner_product(


		vars_.begin(), vars_.end(), sg_->get_var_weights(),	0.0);


	TraceScope("LR",8);


	if(TraceActive()){


		if(target < 0 ) 


			fprintf(stderr, "target cost : %e\n",target);


	}


	


	double fd = std::inner_product(


		delays_.begin(), delays_.end(), mps_.begin(),0.0);


	


	if(TraceActive()){


		if(fd < 0 )


			fprintf(stderr, "fd cost : %e\n",fd);


	}





	return target+fd;


}





double sizing_vars::compute_mp_cost() const


{


	assert(sg_ != 0);


	double fd = std::inner_product(delays_.begin(),delays_.end(), mps_.begin(),0.0);


	return fd;


}





size_t sizing_vars::solve_LRS()


{


	assert(sg_ != 0);





	TraceScope("LRS", 8);





	pre_solve_LRS();





	const size_t N = num_vars();





	std::vector<char> inQ(N, 1);


	std::list<size_t> Q(N);


	for (size_t i = 0; i < N; ++i)


		Q.push_back(i);





	size_t iter = 0;





	for (; !Q.empty(); ++iter)


	{


		size_t v = Q.front();


		Q.pop_front();


		inQ[v] = 0;





		if (!LRS_update_var(v))


			continue;





		typedef sizing_graph::id_crange id_crange;


		for (id_crange cr = sg_->get_related_edges(v);


			cr.first != cr.second; ++cr.first)


			for (id_crange cer = sg_->get_related_vars(*cr.first);


				cer.first != cer.second; ++cer.first)


			{


				size_t u = *cer.first;





				if ((u != v) && (inQ[u] == 0))


				{


					Q.push_back(u);


					inQ[u] = 1;


				}


			}


	}





	if (TraceActive())


		dump(stdout);





	post_solve_LRS();





	return iter;


}





void sizing_vars::pre_solve_LRS()


{


}





void sizing_vars::post_solve_LRS()


{


	update_delays();


	update_arrival_times(0);


}





double sizing_vars::get_achieved_delay()


{


	return ats_[po_];


}





std::pair<double, double> sizing_vars::compute_coef_1m1(size_t which_var)


{


	double coef_1 = sg_->get_var_weight(which_var);


	double coef_m1 = 0;





	for (sizing_graph::id_crange cr = sg_->get_related_edges(which_var);


		cr.first != cr.second; ++cr.first)


	{


		size_t e = *cr.first;


		const math::posynomial &dp = sg_->get_edge_delay(e);





		for (math::posynomial::sum_term_crange ctr = dp.get_term_range();


			ctr.first != ctr.second; ++ctr.first)


		{


			std::pair<double, int> result


				= ctr.first->partial_evaluate(get_vars(), which_var);





			switch (result.second)


			{


			case 0:


				break;


			case 1:


				coef_1 += mps_[e]*result.first;


				break;


			case -1:


				coef_m1 += mps_[e]*result.first;


				break;


			default:


				fprintf(stderr, "error: wrong order of var %u\n", which_var);


				exit(0);


			}


		}


	}





	return std::make_pair(coef_1, coef_m1);


}





bool sizing_vars::LRS_update_var(size_t which_var)


{


	TraceScope("LRS", 9);





	std::pair<double, double> coefs = compute_coef_1m1(which_var);





	double coef_1 = coefs.first;


	double coef_m1 = coefs.second;


	double value = sqrt(coef_m1/coef_1);





	if (fabs(coef_1) < 1e-6)


		value = sg_->get_var_ub(which_var);


	else if (coef_m1 < 0)


		value = sg_->get_var_lb(which_var);





	int action = update_var(which_var, value);





	misc::cstr_t actions[] = {"no change", "updated", "set min", "set max"};





	TracePrint(		


		"update var %u: %s, sqrt(%e/%e)->%e, %s\n",


		which_var,


		sg_->get_var_name(which_var),


		coef_m1,


		coef_1,


		value,


		actions[action]);





	return action != NO_CHANGE;


}





size_t sizing_vars::solve_LRS2()


{


	assert(sg_ != 0);





	TraceScope("LRS", 8);





	pre_solve_LRS();





	size_t iter = 0;


	


	for (;; ++iter)


	{


		bool update_one = false;


		for (size_t i = 0; i < num_vars(); ++i)


			if (LRS_update_var(i))


				update_one = true;





		if (!update_one)


			break;


	}





	if (TraceActive())


		dump(stdout);





	post_solve_LRS();





	return (iter+1)*num_vars();


}





size_t sizing_vars::num_edges() const


{


	assert(sg_ != 0);





	return sg_->num_edges();


}





size_t sizing_vars::num_vars() const


{


	assert(sg_ != 0);





	return sg_->num_vars();


}





size_t sizing_vars::num_vertices() const


{


	assert(sg_ != 0);





	return sg_->num_vertices();


}





double sizing_vars::get_delay(size_t which_edge) const


{


	assert(which_edge < num_edges());





	return delays_[which_edge];


}





double sizing_vars::get_mp(size_t which_edge) const


{


	assert(which_edge < num_edges());





	return mps_[which_edge];


}





double sizing_vars::get_var(size_t which_var) const


{


	assert(which_var < num_vars());





	return vars_[which_var];


}





const sizing_graph *sizing_vars::get_sizing_graph() const


{


	return sg_;


}





void sizing_vars::dump(FILE *fp) const


{


	fprintf(fp, "+++++++++++++++++++++++++++++++++++++++++++++\n");


	for (size_t i = 0; i < num_vars(); ++i)


	{


		fprintf(fp,


			"var %u: %s, %e, [%e, %e]\n",


			i,


			sg_->get_var_name(i),


			vars_[i],


			sg_->get_var_lb(i),


			sg_->get_var_ub(i));


	}


	for (size_t i = 0; i < num_edges(); ++i)


	{


		if (mps_[i] < TOL)


			continue;





		fprintf(fp,


			"edge %i, mp %e, delay %e\n",


			i,


			mps_[i],


			delays_[i]);


	}


	fprintf(fp, "---------------------------------------------\n");


}





void sizing_vars_elmore::attach_graph(const sizing_graph *p_sg)


{


	sizing_vars::attach_graph(p_sg);





	build_coef_caches();


}





int sizing_vars_elmore::update_var(size_t which_var, double value)


{


	int ret = sizing_vars::update_var(which_var, value);





	inv_vars_[which_var] = 1.0/vars_[which_var];





	return ret;


}





void sizing_vars_elmore::pre_solve_LRS()


{


	pre_compute_coef();


}





std::pair<double, double> sizing_vars_elmore::compute_coef_1m1(size_t which_var)


{


	double coef_1 = coef_1s_[which_var].evaluate(get_inv_vars());


	coef_1 += sg_->get_var_weight(which_var);





	double coef_m1 = coef_m1s_[which_var].evaluate(get_vars());





	return std::make_pair(coef_1, coef_m1);


}





void sizing_vars_elmore::build_coef_caches()


{


	inv_vars_.resize(num_vars());


	coef_1s_.resize(num_vars());


	coef_m1s_.resize(num_vars());





	for (size_t i = 0 ;i < num_vars(); ++i)


		inv_vars_[i] = 1.0/vars_[i];





	for (size_t e = 0; e < sg_->num_edges(); ++e)


	{


		const math::posynomial &dp = sg_->get_edge_delay(e);





		for (math::posynomial::sum_term_crange r = dp.get_term_range();


			r.first != r.second; ++r.first)


		{


			const math::posynomial::sum_term & st = *r.first;





			if (st.prod_terms.size() == 0)


			{


				continue;


			}


			else if (st.prod_terms.size() == 1)


			{


				size_t which_var = st.prod_terms[0].var_name;


				int pow_coef = st.prod_terms[0].pow_coef;


				if (pow_coef == -1)


				{


					coef_m1s_[which_var].add_const(st.sum_coef, e);


					continue;


				}


				else if (pow_coef == 1)


				{


					coef_1s_[which_var].add_const(st.sum_coef, e);


					continue;


				}


			}


			else if (st.prod_terms.size() == 2)


			{


				size_t which_var = st.prod_terms[0].var_name;


				int pow_coef = st.prod_terms[0].pow_coef;


				size_t other_var = st.prod_terms[1].var_name;


				int other_pow = st.prod_terms[1].pow_coef;





				if (pow_coef != 1)


				{


					std::swap(pow_coef, other_pow);


					std::swap(which_var, other_var);


				}





				if ((pow_coef == 1) && (other_pow == -1))


				{


					coef_1s_[which_var].add_coef(st.sum_coef, other_var, e);


					coef_m1s_[other_var].add_coef(st.sum_coef, which_var, e);


					continue;


				}


			}





			fprintf(stderr, "error: not a proper Elmore delay for edge %u\n", e);


			exit(0);


		}


	}


}





void sizing_vars_elmore::pre_compute_coef()


{


	for (size_t i = 0; i < num_vars(); ++i)


	{


		coef_1s_[i].update_cache(get_mps());


		coef_m1s_[i].update_cache(get_mps());


	}


}





void sizing_vars_elmore::coef_cache::add_const(double coef, size_t which_edge)


{


	const_mps_.push_back(std::make_pair(coef, which_edge));


}





void sizing_vars_elmore::coef_cache::add_coef(


	double coef, size_t which_var, size_t which_edge)


{


	coef_mps_.push_back(std::make_pair(coef, which_edge));


	coef_vars_.push_back(std::make_pair(0.0, which_var));


}





void sizing_vars_elmore::coef_cache::update_cache(const double mps[])


{


	cached_const_ = 0;


	for (size_t i = 0; i < const_mps_.size(); ++i)


		cached_const_ += const_mps_[i].first*mps[const_mps_[i].second];





	assert(coef_vars_.size() == coef_mps_.size());


	for (size_t i = 0; i < coef_mps_.size(); ++i)


		coef_vars_[i].first = coef_mps_[i].first*mps[coef_mps_[i].second];


}





double sizing_vars_elmore::coef_cache::evaluate(const double vars[])


{


	double ret = cached_const_;





	for (size_t i = 0; i < coef_vars_.size(); ++i)


		ret += coef_vars_[i].first*vars[coef_vars_[i].second];





	return ret;


}





const double *sizing_vars_elmore::get_inv_vars() const


{


	assert(sg_ != 0);





	return &inv_vars_[0];


}


