#ifndef SIZING_VARS_H
#define SIZING_VARS_H

#include "sizing_graph.h"

class sizing_vars
{
public:
	sizing_vars();

	virtual void attach_graph(const sizing_graph *p_sg);
	const sizing_graph *get_sizing_graph() const;

	const double *get_vars() const;
	const double *get_delays() const;

	const double *get_mps() const;

	size_t num_edges() const;
	size_t num_vars() const;
	size_t num_vertices() const;

	double get_var(size_t which_edge) const;
	double get_delay(size_t which_edge) const;
	
	double get_mp(size_t which_edge) const;
	virtual void set_mp(size_t which_edge, double mp);
	//For traditional LR sizing
	double set_mp_slack(size_t which_edge, double step, double norm);
	double set_mp_slack_nostep(size_t which_edge);
	void project_multipliers(int iter);
	void update_arrival_times(int iter);
	void identify_po_pi();
	double get_achieved_delay();

	void dump(FILE *fp) const;

	virtual size_t solve_LRS();
	virtual size_t solve_LRS2();

	double get_target() const;
	double get_fd() const;
	double get_cost() const;
	double get_cost_ub() const;

	double compute_cost() const;
	double compute_mp_cost() const;
	double compute_subgradient_norm() const;

protected:
	// will decide where to go and update the cache if needed
	enum {NO_CHANGE = 0, UPDATED, SET_MIN, SET_MAX};
	virtual int update_var(size_t which_var, double value);

	void initialize_vars();
	void initialize_ats();

	void update_delays();
	bool LRS_update_var(size_t which_var);

	virtual void pre_solve_LRS();
	virtual void post_solve_LRS();
	virtual std::pair<double, double> compute_coef_1m1(size_t which_var);

	const sizing_graph *sg_;
	
	std::vector<double> vars_, delays_, mps_;
	//arrival time updates
	std::vector<double> ats_,dats_;
	//source vertices for bfs to do arrival time updates
	std::vector<size_t> sources_;
	double target_, fd_, cost_ub_;
	//identify PO to PI edge 
	//potentials on this edge cannot change as they represent the timing constraints
	size_t po_,pi_;

}; // class sizing_vars

// speed up coef_1/coef_m1 computation in sizing_vars::LRS_update_var
// for Elmore delays by caching
class sizing_vars_elmore : public sizing_vars
{
public:
	virtual void attach_graph(const sizing_graph *p_sg);

	const double *get_inv_vars() const;

protected:
	virtual int update_var(size_t which_var, double value);
	virtual void pre_solve_LRS();
	virtual std::pair<double, double> compute_coef_1m1(size_t which_var);

	void build_coef_caches();
	void pre_compute_coef();

	std::vector<double> inv_vars_;

	class coef_cache
	{
	public:
		typedef std::pair<double, size_t> coef_index_pair;

		void add_const(double coef, size_t which_edge);
		void add_coef(double coef, size_t which_var, size_t which_edge);

		void update_cache(const double mps[]);

		double evaluate(const double vars[]);

	protected:
		double cached_const_;

		std::vector<coef_index_pair> const_mps_;
		std::vector<coef_index_pair> coef_vars_;
		std::vector<coef_index_pair> coef_mps_;

	}; // class coef_cache

	std::vector<coef_cache> coef_1s_, coef_m1s_;

}; // class sizing_vars_elmore

#endif // SIZING_VARS_H
