#include "../misc/common.h"
#include "../misc/trace.h"

#include "sizing_vars.h"

void sizing_vars_elmore::attach_graph(const sizing_graph *p_sg)
{
	sizing_vars::attach_graph(p_sg);

	build_coef_caches();
}

int sizing_vars_elmore::update_var(size_t which_var, double value)
{
	int ret = sizing_vars::update_var(which_var, value);

	inv_vars_[which_var] = 1.0/vars_[which_var];

	return ret;
}

void sizing_vars_elmore::pre_solve_LRS()
{
	pre_compute_coef();
}

std::pair<double, double> sizing_vars_elmore::compute_coef_1m1(size_t which_var)
{
	double coef_1 = coef_1s_[which_var].evaluate(get_inv_vars());
	coef_1 += sg_->get_var_weight(which_var);

	double coef_m1 = coef_m1s_[which_var].evaluate(get_vars());

	return std::make_pair(coef_1, coef_m1);
}

void sizing_vars_elmore::build_coef_caches()
{
	inv_vars_.resize(num_vars());
	coef_1s_.resize(num_vars());
	coef_m1s_.resize(num_vars());

	for (size_t i = 0 ;i < num_vars(); ++i)
		inv_vars_[i] = 1.0/vars_[i];

	for (size_t e = 0; e < sg_->num_edges(); ++e)
	{
		const math::posynomial &dp = sg_->get_edge_delay(e);

		for (math::posynomial::sum_term_crange r = dp.get_term_range();
			r.first != r.second; ++r.first)
		{
			const math::posynomial::sum_term & st = *r.first;

			if (st.prod_terms.size() == 0)
			{
				continue;
			}
			else if (st.prod_terms.size() == 1)
			{
				size_t which_var = st.prod_terms[0].var_name;
				int pow_coef = st.prod_terms[0].pow_coef;
				if (pow_coef == -1)
				{
					coef_m1s_[which_var].add_const(st.sum_coef, e);
					continue;
				}
				else if (pow_coef == 1)
				{
					coef_1s_[which_var].add_const(st.sum_coef, e);
					continue;
				}
			}
			else if (st.prod_terms.size() == 2)
			{
				size_t which_var = st.prod_terms[0].var_name;
				int pow_coef = st.prod_terms[0].pow_coef;
				size_t other_var = st.prod_terms[1].var_name;
				int other_pow = st.prod_terms[1].pow_coef;

				if (pow_coef != 1)
				{
					std::swap(pow_coef, other_pow);
					std::swap(which_var, other_var);
				}

				if ((pow_coef == 1) && (other_pow == -1))
				{
					coef_1s_[which_var].add_coef(st.sum_coef, other_var, e);
					coef_m1s_[other_var].add_coef(st.sum_coef, which_var, e);
					continue;
				}
			}

			fprintf(stderr,
				"%s: error: not a proper Elmore delay for edge %lu (%lu->%lu)\n",
				sg_->get_graph_name().c_str(), e,
				sg_->get_edge_from(e), sg_->get_edge_to(e));

			exit(0);
		}
	}
}

void sizing_vars_elmore::pre_compute_coef()
{
	for (size_t i = 0; i < num_vars(); ++i)
	{
		coef_1s_[i].update_cache(get_mps());
		coef_m1s_[i].update_cache(get_mps());
	}
}

void sizing_vars_elmore::coef_cache::add_const(double coef, size_t which_edge)
{
	const_mps_.push_back(std::make_pair(coef, which_edge));
}

void sizing_vars_elmore::coef_cache::add_coef(
	double coef, size_t which_var, size_t which_edge)
{
	coef_mps_.push_back(std::make_pair(coef, which_edge));
	coef_vars_.push_back(std::make_pair(0.0, which_var));
}

void sizing_vars_elmore::coef_cache::update_cache(const double mps[])
{
	cached_const_ = 0;
	for (size_t i = 0; i < const_mps_.size(); ++i)
		cached_const_ += const_mps_[i].first*mps[const_mps_[i].second];

	assert(coef_vars_.size() == coef_mps_.size());
	for (size_t i = 0; i < coef_mps_.size(); ++i)
		coef_vars_[i].first = coef_mps_[i].first*mps[coef_mps_[i].second];
}

double sizing_vars_elmore::coef_cache::evaluate(const double vars[])
{
	double ret = cached_const_;

	for (size_t i = 0; i < coef_vars_.size(); ++i)
		ret += coef_vars_[i].first*vars[coef_vars_[i].second];

	return ret;
}

const double *sizing_vars_elmore::get_inv_vars() const
{
	assert(sg_ != 0);

	return &inv_vars_[0];
}
