/*
 * Copyright (c) 2005-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * fib_heap.h - Fibonacci Heap
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#ifndef FIB_HEAP_H
#define FIB_HEAP_H

#include <cmath>

namespace graph
{

// min fib heap, see CLRS for details
template <class Key, class Value, class Less = std::less<Key> >
class fib_heap
{
	struct node_t
	{
		Key key;
		Value value;
		int degree;
		bool mark;
		int parent, child, left, right;

		node_t(const Key &k, const Value &v, int me)
			: key(k), value(v), degree(0), mark(false),
			parent(-1), child(-1), left(me), right(me)
			{}

		node_t()
			: key(), value(), degree(0), mark(false),
			parent(-1), child(-1), left(-1), right(-1)
			{}

		void reinit(const Key &k, const Value &v)
		{
			key = k;
			value = v;
			degree = 0;
			mark = false;
			parent = -1;
			child = -1;
		}

		void reset(int next_free)
		{
			degree = -1;
			parent = next_free;
		}

		bool inuse() const {return degree != -1;}
	}; // struct node_t

	Less less_;
	std::vector<node_t> pool_;
	int min_, free_;
	size_t size_;

	int alloc(const Key &k, const Value &v)
	{
		if (free_ == -1)
		{
			int me = (int)pool_.size();
			pool_.push_back(node_t(k, v, me));
			return me;
		}
		else
		{
			int me = free_;
			free_ = pool_[free_].parent;
			pool_[me].reinit(k, v);
			return me;
		}
	}
	void free(int i)
	{
		pool_[i].reset(free_);
		free_ = (int)i;
	}

public:
	typedef fib_heap<Key, Value, Less> this_type;
	typedef Value value_type;
	typedef Key key_type;

	fib_heap(Less l = Less())
		: less_(l), min_(-1), free_(-1), size_(0)
		{}

	void reserve(size_t n) {pool_.reserve(n);}
	void clear() {min_ = free_ = -1; size_ = 0; pool_.resize(0);}
	size_t size() const {return size_;}
	bool empty() const {return size_ == 0;}

	bool less(const key_type &a, const key_type &b) const {return less_(a, b);}

	// min
	int get_min() const {return min_;}

	// value
	const value_type &value(int i) const
		{assert(valid_index(i)); return pool_[i].value;}
	value_type &value(int i)
		{assert(valid_index(i)); return pool_[i].value;}

	// key
	const key_type &key(int i) const
		{assert(valid_index(i)); return pool_[i].key;}
	
	// heap operations
	void extract_min()
	{
		assert(!empty());

		--size_;

		// put children of min into root list
		int child = pool_[min_].child;
		if (child != -1)
		{
			// update parent
			for (; pool_[child].parent == min_;
				child = pool_[child].right)
				pool_[child].parent = -1;

			// merge to root list
			list_insert(child, min_);
		}

		// erase min from the list
		int temp = min_;
		min_ = list_unlink(min_);
		free(temp);

		if (size_ != 0)
			consolidate();
	}

	void decrease_key(int x, const key_type &new_k)
	{
		assert(less_(new_k, key(x)));

		pool_[x].key = new_k;

		int y = pool_[x].parent;
		if ((y != -1) && less_(key(x), key(y)))
		{
			cut(x, y);
			cascading_cut(y);
		}

		// see if we need to update min
		if (less_(key(x), key(min_)))
			min_ = x;
	}

	int insert(const key_type &k, const value_type &v)
	{
		++size_;

		int me = alloc(k, v);
		
		// add to empty heap
		if (min_ == -1)
		{
			min_ = me;
			return me;
		}
		
		// insert to root list
		list_insert(me, min_);

		// see if we need to update min
		if (less_(key(me), key(min_)))
			min_ = me;

		return me;
	}
	
	void erase(int i)
	{
		assert(valid_index(i));

		// decrease to -infinity
		int y = pool_[i].parent;
		if (y != -1)
		{
			cut(i, y);
			cascading_cut(y);
		}

		min_ = i;

		extract_min();
	}

private:

	bool valid_index(int i) const
	{
		return (size_t(i) < pool_.size()) && pool_[i].inuse();
	}

	// insert before after
	void list_insert(int i, int after)
	{
		assert(valid_index(after));
		assert(valid_index(i));

		// l<->m
		// l<->i2<->...<->i1<->m
		node_t &m = pool_[after];
		node_t &l = pool_[m.left];
		node_t &i1 = pool_[i];
		node_t &i2 = pool_[i1.right];

		i2.left = m.left;
		m.left = i;
		l.right = i1.right;
		i1.right = after;
	}

	int list_unlink(int i)
	{
		assert(valid_index(i));

		int l = pool_[i].left;
		int r = pool_[i].right;

		pool_[i].left = pool_[i].right = i;

		if (i != r)
		{		
			pool_[r].left = l;
			pool_[l].right = r;
			return r;
		}
		else
			return -1;
	}

	void list_erase_child(int i)
	{
		assert(valid_index(i));

		int parent = pool_[i].parent;

		if (parent == -1)
			list_unlink(i);
		else
		{
			pool_[i].parent = -1;
			pool_[parent].child = list_unlink(i);
			--pool_[parent].degree;
		}
	}

	void list_add_child(int i, int parent)
	{
		assert(valid_index(parent));
		assert(valid_index(i));

		node_t &p = pool_[parent];
		
		if (p.child == -1)
			p.child = i;
		else
			list_insert(i, p.child);

		++p.degree;
		pool_[i].parent = parent;
	}

	void consolidate()
	{
		assert(min_ != -1);

		int A[64];

		const size_t N = size_t(log(double(size_))/0.4811)+1;

		std::fill(A, A+N, -1);

		for (int cur = min_, last = pool_[min_].left;;)
		{
			int next = pool_[cur].right;

			int x = cur;
			int d = pool_[x].degree;

			for (; A[d] != -1; A[d++] = -1)
			{
				int y = A[d];

				if (less_(key(y), key(x)))
					std::swap(x, y);

				list_erase_child(y);
				list_add_child(y, x);
				pool_[y].mark = false;
			}
			A[d] = x;

			if (cur == last)
				break;
			else
				cur = next;
		}

		min_ = -1;
		for (size_t i = 0; i < N; ++i)
			if ((A[i] != -1) && ((min_ == -1)
				|| less_(key(A[i]), key(min_))))
				min_ = A[i];
	}

	void cut(int x, int y)
	{
		assert(valid_index(x));
		assert(valid_index(y));
		assert(pool_[x].parent == y);

		list_erase_child(x);
		list_insert(x, min_);
		pool_[x].mark = false;
	}

	void cascading_cut(int y)
	{
		assert(valid_index(y));

		node_t &n = pool_[y];
		int p = n.parent;
		if (p != -1)
		{
			if (!n.mark)
			{
				n.mark = true;
			}
			else
			{
				cut(y, p);
				cascading_cut(p);
			}
		}
	}

}; // class fib_heap<Key, Value, Less>

} // namespace graph

#endif // FIB_HEAP_H
