/*
 * Copyright (c) 2003-2005 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * graph2_vertex.h - Interface for vertex management.
 *
 * Authors: Jia Wang, jwa112@ece.northwestern.edu
 *
 */

#ifndef GRAPH2_VERTEX_H
#define GRAPH2_VERTEX_H

#include "graph2_property.h"

namespace graph2
{

namespace detail
{

class vertex_descriptor_impl
{
	size_t v_;
public:
	explicit vertex_descriptor_impl(size_t v) : v_(v) {}
	vertex_descriptor_impl() : v_(size_t(-1)) {}
	size_t id() const {return v_;}

	bool operator == (vertex_descriptor_impl v) const {return v_ == v.v_;}

	bool operator < (vertex_descriptor_impl r) const {return v_ < r.v_;}
	bool operator > (vertex_descriptor_impl r) const {return v_ > r.v_;}

	// use valid_vertex() as an end() iterator
	vertex_descriptor_impl& operator ++ () {++v_; return *this;}
	vertex_descriptor_impl& operator -- () {--v_; return *this;}
}; // class vertex_descriptor_impl

class vertex_set_base
{
protected:
	size_t num_vertices_;

	vertex_set_base(size_t n = 0) : num_vertices_(n) {}

public:
	typedef vertex_descriptor_impl vertex_descriptor;

	size_t num_vertices() const {return num_vertices_;}
	bool valid_vertex(vertex_descriptor v) const {return v.id() < num_vertices();}

	// we need a concrete type for vertex_vector_property
	template <class T>
	struct vertex_vector_property : public vector_property<vertex_descriptor, T>
	{
		vertex_vector_property(const vertex_set_base &V, const T &def = T())
			: vector_property<vertex_descriptor, T>(V.num_vertices(), def)
			{}

		template <class Iter>
		vertex_vector_property(const vertex_set_base &V, Iter begin, Iter end)
			: vector_property<vertex_descriptor, T>(V.num_vertices(), begin, end)
			{}
	}; // struct vertex_vector_property<T>

	// shortcuts are enough for const and array properties
	template <class T>
	const_property<vertex_descriptor, T>
	make_vertex_property(T c) const
	{
		return const_property<vertex_descriptor, T>(c);
	}

	template <class T>
	array_property<vertex_descriptor, T>
	make_vertex_property(T props[], size_t n) const
	{
		assert(num_vertices() <= n);
		return array_property<vertex_descriptor, T>(props, n);
	}

	template <class T>
	array_const_property<vertex_descriptor, T>
	make_vertex_property(const T props[], size_t n) const
	{
		assert(num_vertices() <= n);
		return array_const_property<vertex_descriptor, T>(props, n);
	}

	template <class T>
	array_property<vertex_descriptor, T>
	make_vertex_property(std::vector<T> &props) const
	{
		assert(num_vertices() <= props.size());
		return array_property<vertex_descriptor, T>(&props[0], props.size());
	}

	template <class T>
	array_const_property<vertex_descriptor, T>
	make_vertex_property(const std::vector<T> &props) const
	{
		assert(num_vertices() <= props.size());
		return array_const_property<vertex_descriptor, T>(&props[0], props.size());
	}

}; // class vertex_set_base

} // namespace detail

} // namespace graph2

#endif // GRAPH2_VERTEX_H
