/*
 * Copyright (c) 2003-2007 Northwestern University.
 *                         NuCAD Group.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * howard.h - howard algorithm for min/max cycle mean
 *
 * Authors: Jia Wang, jwa112@eecs.northwestern.edu
 *
 */

#ifndef HOWARD_H
#define HOWARD_H

#include "graph2_v.h"

namespace graph2
{

template <class Weight, class Length>
class howard
{
	const graph_v &graph_;

	Weight weight_;
	Length length_;

	graph_v::vertex_vector_property<double> d_;
	graph_v::vertex_vector_property<graph_v::edge_descriptor> pi_;

	// need to track the pi graph :(
	// use a dynamic graph for pi next time
	typedef std::vector<graph_v::vertex_descriptor> back_v_t;
	graph_v::vertex_vector_property<back_v_t> pi_back_;

	// for pi_min_cycle_mean
	graph_v::vertex_vector_property<double> path_length_, path_weight_;
	graph_v::vertex_vector_property<int> branchs_;

	// for bfs_update_d
	graph_v::vertex_vector_property<char> visited_;
	std::vector<graph_v::vertex_descriptor> queue_;

	static void insert(back_v_t &back, graph_v::vertex_descriptor v)
	{
		assert(std::find(back.begin(), back.end(), v) == back.end());
		back.push_back(v);
	}

	static void erase(back_v_t &back, graph_v::vertex_descriptor v)
	{
		back_v_t::iterator it = std::find(back.begin(), back.end(), v);
		assert(it != back.end());
		*it = back.back();
		back.pop_back();
	}

	bool inited_;
	int count_;

	void init()
	{
		if (!inited_)
		{
			d_.fill(1e100);

			for (graph_v::edge_descriptor e(0); graph_.valid_edge(e); ++e)
			{
				graph_v::vertex_descriptor u = graph_.source(e);
				graph_v::vertex_descriptor v = graph_.target(e);
				if (weight_[e] < d_[u])
				{
					if (graph_.valid_edge(pi_[u])) // remove previous pi edge first
						erase(pi_back_[graph_.target(pi_[u])], u);
					d_[u] = weight_[e];
					pi_[u] = e;
					insert(pi_back_[v], u);
				}
			}

			inited_ = true;
		}
		else
		{
			for (graph_v::vertex_descriptor v(0); graph_.valid_vertex(v); ++v)
				d_[v] = weight_[pi_[v]];
		}
	}

	std::pair<double, graph_v::vertex_descriptor> pi_min_cycle_mean()
	{
		// find cycle in pi, every vertex belong to at most one cycle
		branchs_.fill(-1);

		double min_mean = 1e100;
		graph_v::vertex_descriptor min_vertex;
		int this_branch = -1;

		for (graph_v::vertex_descriptor v(0); graph_.valid_vertex(v); ++v)
			if (branchs_[v] == -1) // reach a new branch
			{
				assert(graph_.valid_edge(pi_[v]));
				
				++this_branch;

				double length = 0, weight = 0;

/* example
 * u_0-1->u_1-2->u_2-3->u_3-4->u_1    -> length = weight = 0 when find u = u_0
 * 1      2      3      4             -> path_length(u)
 * 1      3      6      10            -> path_weight(u)
 */
				for (graph_v::vertex_descriptor u = v;;)
				{
					path_length_[u] = (length += length_[pi_[u]]);
					path_weight_[u] = (weight += weight_[pi_[u]]);
					branchs_[u] = this_branch;

//					fprintf(stdout,
//						"HOWARD: info: found %u in branch %d, %.1f\n",
//						u.id(), this_branch, length);

					u = graph_.target(pi_[u]);
					assert(graph_.valid_edge(pi_[u]));

                    if (branchs_[u] == -1) // reach an new vertex
						continue;
					else if (branchs_[u] < this_branch) // reach an old path
					{
//						fprintf(stdout,
//							"HOWARD: info: lead to %u in branch %d\n",
//							u.id(), branchs_[u]);
			
						break;
					}
					else // find a new cycle started at u
					{
						length += length_[pi_[u]];
						weight += weight_[pi_[u]];
						double mean = (weight-path_weight_[u])/(length-path_length_[u]);

						fprintf(stdout,
							"HOWARD: info: cycle in branch %d, %f\n",
							this_branch, mean);

						if (mean < min_mean)
							min_mean = mean, min_vertex = u;
						break;
					}
				}
			}
			return std::make_pair(min_mean, min_vertex);
	}

	void bfs_update_d(graph_v::vertex_descriptor s, double mean)
	{
		visited_.fill(0);
		size_t qb = 0, qe = 0;

		visited_[s] = 1;
		for (back_v_t::iterator ib = pi_back_[s].begin();
			ib != pi_back_[s].end(); ++ib)
			queue_[qe++] = *ib;

		while (qb != qe)
		{
			graph_v::vertex_descriptor u = queue_[qb++];
			visited_[u] = 1;

			graph_v::edge_descriptor e = pi_[u];

			graph_v::vertex_descriptor v = graph_.target(e);
			assert(visited_[v]);
			d_[u] = d_[v]+weight_[e]-mean*length_[e];
			
			for (back_v_t::iterator ib = pi_back_[u].begin();
				ib != pi_back_[u].end(); ++ib)
				if (visited_[*ib] == 0)
					queue_[qe++] = *ib;
		}
	}

	bool improve_d(double mean)
	{
		bool ret = false;

		for (graph_v::edge_descriptor e(0); graph_.valid_edge(e); ++e)
		{
			graph_v::vertex_descriptor u = graph_.source(e);
			graph_v::vertex_descriptor v = graph_.target(e);

			double new_du = d_[v]+weight_[e]-mean*length_[e];
			double delta = d_[u]-new_du;
			if (delta > 0)
			{
				if (delta > 0.001)
				{
					ret = true;
					if (graph_.valid_edge(pi_[u])) // remove previous pi edge first
						erase(pi_back_[graph_.target(pi_[u])], u);
					d_[u] = new_du;
					pi_[u] = e;
					insert(pi_back_[v], u);
				}
			}
		}
		return ret;
	}

public:
	howard(const graph_v &g, Weight weight, Length length)
		: graph_(g), weight_(weight), length_(length),
		d_(g), pi_(g), pi_back_(g),
		path_length_(g), path_weight_(g), branchs_(g),
		visited_(g), queue_(g.num_vertices()),
		inited_(false), count_(0)
	{
		for (graph_v::vertex_descriptor v(0); graph_.valid_vertex(v); ++v)
			pi_back_[v].reserve(16);
	}

	~howard()
	{
//		fprintf(stderr, "Howard %d: %d\n", queue_.size(), count_);
	}

	int count() const {return count_;}

	double apply()
	{
		init();
		for (;;)
		{
			++count_;
			std::pair<double, graph_v::vertex_descriptor>
				min_cycle = pi_min_cycle_mean();
			fprintf(stdout, "HOWARD: info: current cycle %lf\n", min_cycle.first);
			bfs_update_d(min_cycle.second, min_cycle.first);
			if (!improve_d(min_cycle.first))
				return min_cycle.first;
		}
	}

	bool BF_verify(double lambda)
	{
		graph_v::vertex_vector_property<double> d(graph_, 0);

		for (size_t i = 0; i < graph_.num_vertices(); ++i)
		{
			for (graph_v::edge_descriptor e(0); graph_.valid_edge(e); ++e)
			{
				graph_v::vertex_descriptor u = graph_.source(e);
				graph_v::vertex_descriptor v = graph_.target(e);

				double bound = d[u]+weight_[e]-length_[e]*lambda;
				if (d[v] > bound)
					d[v] = bound;
			}
		}
		for (graph_v::edge_descriptor e(0); graph_.valid_edge(e); ++e)
		{
			graph_v::vertex_descriptor u = graph_.source(e);
			graph_v::vertex_descriptor v = graph_.target(e);

			double bound = d[u]+weight_[e]-length_[e]*lambda;
			if (d[v] > bound)
				return false;
		}

		return true;
	}

}; // class howard<Weight, Length>

} // namespace graph2

#endif // HOWARD_H
